"""This file contains a plugin for parsing Google Analytics cookies."""

from urllib import parse as urlparse

from plaso.containers import events
from plaso.parsers.cookie_plugins import interface
from plaso.parsers.cookie_plugins import manager


# TODO: determine if __utmc always 0?


class GoogleAnalyticsUtmaEventData(events.EventData):
  """Google analytics __utma cookie event data.

  Attributes:
    cookie_name (str): name of cookie.
    domain_hash (str): domain hash.
    sessions (int): number of sessions.
    url (str): URL or path where the cookie got set.
    visited_times (list[dfdatetime.DateTimeValues]): dates and times the URL
        was visited.
    visitor_identifier (str): visitor identifier.
  """

  DATA_TYPE = 'cookie:google:analytics:utma'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.cookie_name = None
    self.domain_hash = None
    self.sessions = None
    self.url = None
    self.visited_times = None
    self.visitor_identifier = None


class GoogleAnalyticsUtmbEventData(events.EventData):
  """Google analytics __utmb cookie event data.

  Attributes:
    cookie_name (str): name of cookie.
    domain_hash (str): domain hash.
    last_visited_time (dfdatetime.DateTimeValues): date and time the URL was
        last visited.
    pages_viewed (int): number of pages viewed.
    url (str): URL or path where the cookie got set.
  """

  DATA_TYPE = 'cookie:google:analytics:utmb'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.cookie_name = None
    self.domain_hash = None
    self.last_visited_time = None
    self.pages_viewed = None
    self.url = None


class GoogleAnalyticsUtmtEventData(events.EventData):
  """Google analytics __utmt cookie event data.

  Attributes:
    cookie_name (str): name of cookie.
    last_visited_time (dfdatetime.DateTimeValues): date and time the URL was
        last visited.
    url (str): URL or path where the cookie got set.
  """

  DATA_TYPE = 'cookie:google:analytics:utmt'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.cookie_name = None
    self.last_visited_time = None
    self.url = None


class GoogleAnalyticsUtmzEventData(events.EventData):
  """Google analytics __utmz cookie event data.

  Attributes:
    cookie_name (str): name of cookie.
    domain_hash (str): domain hash.
    last_visited_time (dfdatetime.DateTimeValues): date and time the URL was
        last visited.
    sessions (int): number of sessions.
    sources (int): number of sources.
    url (str): URL or path where the cookie got set.
  """

  DATA_TYPE = 'cookie:google:analytics:utmz'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.cookie_name = None
    self.domain_hash = None
    self.last_visited_time = None
    self.sessions = None
    self.sources = None
    self.url = None


class GoogleAnalyticsUtmaPlugin(interface.BaseCookiePlugin):
  """A browser cookie plugin for __utma Google Analytics cookies.

  The structure of the cookie data:
  <domain hash>.<visitor ID>.<first visit>.<previous visit>.<last visit>.
  <number of sessions>

  For example:
  137167072.1215918423.1383170166.1383170166.1383170166.1

  Or:
  <last visit>

  For example:
  13113225820000000
  """

  NAME = 'google_analytics_utma'
  DATA_FORMAT = 'Google Analytics __utma cookie'

  COOKIE_NAME = '__utma'

  def _ParseCookieData(
      self, parser_mediator, cookie_data=None, url=None, **kwargs):
    """Extracts events from cookie data.

    Args:
      parser_mediator (ParserMediator): parser mediator.
      cookie_data (str): cookie data.
      url (str): URL or path where the cookie got set.
    """
    fields = cookie_data.split('.')
    number_of_fields = len(fields)

    if number_of_fields not in (1, 6):
      parser_mediator.ProduceExtractionWarning(
          f'unsupported number of fields: {number_of_fields:d} in cookie: '
          f'{self.COOKIE_NAME:s}')
      return

    visited_times = []

    if number_of_fields == 1:
      domain_hash = None
      visitor_identifier = None
      number_of_sessions = None

      date_time = self._ParsePosixTimeIn100Nanoseconds(fields[0])
      if date_time:
        visited_times.append(date_time)

    elif number_of_fields == 6:
      domain_hash = fields[0]
      visitor_identifier = fields[1]

      # TODO: Double check this time is stored in UTC and not local time.
      date_time = self._ParsePosixTime(fields[2])
      if date_time:
        visited_times.append(date_time)

      date_time = self._ParsePosixTime(fields[3])
      if date_time:
        visited_times.append(date_time)

      date_time = self._ParsePosixTime(fields[4])
      if date_time:
        visited_times.append(date_time)

      number_of_sessions = self._ParseIntegerValue(fields[5])

    else:
      domain_hash = None
      number_of_sessions = None
      visitor_identifier = None

    event_data = GoogleAnalyticsUtmaEventData()
    event_data.cookie_name = self.COOKIE_NAME
    event_data.domain_hash = domain_hash
    event_data.sessions = number_of_sessions
    event_data.url = url
    event_data.visited_times = visited_times or None
    event_data.visitor_identifier = visitor_identifier

    parser_mediator.ProduceEventData(event_data)


class GoogleAnalyticsUtmbPlugin(interface.BaseCookiePlugin):
  """A browser cookie plugin for __utmb Google Analytics cookies.

  The structure of the cookie data:
  <domain hash>.<pages viewed>.<unknown>.<last time>

  For example:
  137167072.1.10.1383170166
  173272373.6.8.1440489514899
  173272373.4.9.1373300660574

  Or:
  <last time>

  For example:
  13113225820000000
  """

  NAME = 'google_analytics_utmb'
  DATA_FORMAT = 'Google Analytics __utmb cookie'

  COOKIE_NAME = '__utmb'

  def _ParseCookieData(
      self, parser_mediator, cookie_data=None, url=None, **kwargs):
    """Extracts events from cookie data.

    Args:
      parser_mediator (ParserMediator): parser mediator.
      cookie_data (bytes): cookie data.
      url (str): URL or path where the cookie got set.
    """
    fields = cookie_data.split('.')
    number_of_fields = len(fields)

    if number_of_fields not in (1, 4):
      parser_mediator.ProduceExtractionWarning(
          f'unsupported number of fields: {number_of_fields:d} in cookie: '
          f'{self.COOKIE_NAME:s}')
      return

    if number_of_fields == 1:
      domain_hash = None
      number_of_pages_viewed = None

      date_time = self._ParsePosixTimeIn100Nanoseconds(fields[0])

    elif number_of_fields == 4:
      domain_hash = fields[0]

      number_of_pages_viewed = self._ParseIntegerValue(fields[1])

      if fields[2] in ('8', '9'):
        date_time = self._ParsePosixTimeInMilliseconds(fields[3])
      else:
        date_time = self._ParsePosixTime(fields[3])

    else:
      date_time = None
      domain_hash = None
      number_of_pages_viewed = None

    event_data = GoogleAnalyticsUtmbEventData()
    event_data.cookie_name = self.COOKIE_NAME
    event_data.domain_hash = domain_hash
    event_data.last_visited_time = date_time
    event_data.pages_viewed = number_of_pages_viewed
    event_data.url = url

    parser_mediator.ProduceEventData(event_data)


class GoogleAnalyticsUtmtPlugin(interface.BaseCookiePlugin):
  """A browser cookie plugin for __utmt Google Analytics cookies.

  The structure of the cookie data:
  <last time>

  For example:
  13113215173000000
  """

  NAME = 'google_analytics_utmt'
  DATA_FORMAT = 'Google Analytics __utmt cookie'

  COOKIE_NAME = '__utmt'

  def _ParseCookieData(
      self, parser_mediator, cookie_data=None, url=None, **kwargs):
    """Extracts events from cookie data.

    Args:
      parser_mediator (ParserMediator): parser mediator.
      cookie_data (bytes): cookie data.
      url (str): URL or path where the cookie got set.
    """
    fields = cookie_data.split('.')
    number_of_fields = len(fields)

    if number_of_fields != 1:
      parser_mediator.ProduceExtractionWarning(
          f'unsupported number of fields: {number_of_fields:d} in cookie: '
          f'{self.COOKIE_NAME:s}')
      return

    event_data = GoogleAnalyticsUtmtEventData()
    event_data.cookie_name = self.COOKIE_NAME
    event_data.last_visited_time = self._ParsePosixTimeIn100Nanoseconds(
        fields[0])
    event_data.url = url

    parser_mediator.ProduceEventData(event_data)


class GoogleAnalyticsUtmzPlugin(interface.BaseCookiePlugin):
  """A browser cookie plugin for __utmz Google Analytics cookies.

  The structure of the cookie data:
  <domain hash>.<last time>.<sessions>.<sources>.<variables>

  For example:
  207318870.1383170190.1.1.utmcsr=google|utmccn=(organic)|utmcmd=organic|
  utmctr=(not%20provided)

  Or:
  <last time>

  For example:
  13128990382000000
  """

  NAME = 'google_analytics_utmz'
  DATA_FORMAT = 'Google Analytics __utmz cookie'

  COOKIE_NAME = '__utmz'

  def _ParseCookieData(
      self, parser_mediator, cookie_data=None, url=None, **kwargs):
    """Extracts events from cookie data.

    Args:
      parser_mediator (ParserMediator): parser mediator.
      cookie_data (str): cookie data.
      url (str): URL or path where the cookie got set.
    """
    fields = cookie_data.split('.')
    number_of_fields = len(fields)

    if number_of_fields > 5:
      variables = '.'.join(fields[4:])
      fields = fields[0:4]
      fields.append(variables)
      number_of_fields = len(fields)

    if number_of_fields not in (1, 5):
      parser_mediator.ProduceExtractionWarning(
          f'unsupported number of fields: {number_of_fields:d} in cookie: '
          f'{self.COOKIE_NAME:s}')
      return

    if number_of_fields == 1:
      domain_hash = None
      number_of_sessions = None
      number_of_sources = None
      extra_attributes = {}

      date_time = self._ParsePosixTimeIn100Nanoseconds(fields[0])

    elif number_of_fields == 5:
      domain_hash = fields[0]

      date_time = self._ParsePosixTime(fields[1])

      number_of_sessions = self._ParseIntegerValue(fields[2])
      number_of_sources = self._ParseIntegerValue(fields[3])

      extra_variables = fields[4].split('|')

      extra_attributes = {}
      for variable in extra_variables:
        key, _, value = variable.partition('=')
        extra_attributes[key] = urlparse.unquote(value)

    else:
      date_time = None
      domain_hash = None
      number_of_sessions = None
      number_of_sources = None

    event_data = GoogleAnalyticsUtmzEventData()
    event_data.cookie_name = self.COOKIE_NAME
    event_data.domain_hash = domain_hash
    event_data.last_visited_time = date_time
    event_data.sessions = number_of_sessions
    event_data.sources = number_of_sources
    event_data.url = url

    # TODO: explicitly define these as attributes of
    # GoogleAnalyticsUtmzEventData.
    for key, value in extra_attributes.items():
      setattr(event_data, key, value)

    parser_mediator.ProduceEventData(event_data)


manager.CookiePluginsManager.RegisterPlugins([
    GoogleAnalyticsUtmaPlugin, GoogleAnalyticsUtmbPlugin,
    GoogleAnalyticsUtmtPlugin, GoogleAnalyticsUtmzPlugin])



"""The psort CLI tool."""

import argparse
import os

# The following import makes sure the filters are registered.
from plaso import filters  # pylint: disable=unused-import

# The following import makes sure the output modules are registered.
from plaso import output   # pylint: disable=unused-import

from plaso.cli import analysis_tool
from plaso.cli import logger
from plaso.cli import status_view
from plaso.cli import tool_options
from plaso.cli import views
from plaso.cli.helpers import manager as helpers_manager
from plaso.containers import reports
from plaso.engine import configurations
from plaso.engine import engine
from plaso.helpers import language_tags
from plaso.lib import definitions
from plaso.lib import errors
from plaso.lib import loggers
from plaso.multi_process import output_engine as multi_output_engine
from plaso.storage import factory as storage_factory


class PsortTool(
    analysis_tool.AnalysisTool,
    tool_options.OutputModuleOptions):
  """Psort CLI tool.

  Attributes:
    list_analysis_plugins (bool): True if information about the analysis
        plugins should be shown.
    list_language_tags (bool): True if the language tags should be listed.
    list_output_modules (bool): True if information about the output modules
        should be shown.
    list_profilers (bool): True if the profilers should be listed.
  """

  NAME = 'psort'
  DESCRIPTION = (
      'Application to read, filter and process output from a Plaso storage '
      'file.')

  _CONTAINER_TYPE_ANALYSIS_REPORT = reports.AnalysisReport.CONTAINER_TYPE

  def __init__(self, input_reader=None, output_writer=None):
    """Initializes the CLI tool object.

    Args:
      input_reader (Optional[InputReader]): input reader, where None indicates
          that the stdin input reader should be used.
      output_writer (Optional[OutputWriter]): output writer, where None
          indicates that the stdout output writer should be used.
    """
    super().__init__(
        input_reader=input_reader, output_writer=output_writer)
    self._deduplicate_events = True
    self._preferred_language = None
    self._process_memory_limit = None
    self._status_view = status_view.StatusView(self._output_writer, self.NAME)
    self._status_view_file = 'status.info'
    self._status_view_mode = status_view.StatusView.MODE_WINDOW
    self._time_slice = None
    self._use_time_slicer = False

    self.list_language_tags = False
    self.list_output_modules = False
    self.list_profilers = False

  def _CheckStorageFile(
      self, storage_file_path, check_readable_only=False):
    """Checks if the storage file is valid.

    Args:
      storage_file_path (str): path of the storage file.
      check_readable_only (Optional[bool]): whether the storage file should
          only be checked to see if it can be read. If False, the store will
          be checked to see if it can be read and written to.

    Raises:
      BadConfigOption: if the storage file is invalid.
    """
    if not storage_file_path:
      raise errors.BadConfigOption('Missing storage file option.')

    if not os.path.exists(storage_file_path):
      raise errors.BadConfigOption(
          f'No such storage file: {storage_file_path:s}')

    if not os.path.isfile(storage_file_path):
      raise errors.BadConfigOption((
          f'Storage file: {storage_file_path:s} already exists and is not '
          f'a file.'))

    if not check_readable_only:
      storage_file_directory = os.path.dirname(storage_file_path) or '.'
      if not os.access(storage_file_directory, os.W_OK):
        raise errors.BadConfigOption(
            f'Unable to write to storage file: {storage_file_path:s}')

    storage_file = storage_factory.StorageFactory.CreateStorageFile(
        definitions.STORAGE_FORMAT_SQLITE)
    if not storage_file:
      raise errors.BadConfigOption(
          f'Unable to open storage file: {storage_file_path:s}')

    try:
      storage_file.Open(
          path=storage_file_path, read_only=check_readable_only)
    except OSError as exception:
      raise errors.BadConfigOption((
          f'Unable to open storage file: {storage_file_path:s} with error: '
          f'{exception!s}'))

    storage_file.Close()

  def _CreateOutputAndFormattingProcessingConfiguration(self):
    """Creates an output and formatting processing configuration.

    Returns:
      ProcessingConfiguration: output and formatting processing configuration.
    """
    configuration = configurations.ProcessingConfiguration()
    configuration.custom_formatters_path = self._output_custom_formatters_path
    configuration.data_location = self._data_location
    configuration.debug_output = self._debug_mode
    configuration.dynamic_time = self._output_dynamic_time
    configuration.log_filename = self._log_file
    configuration.preferred_language = self._preferred_language
    configuration.preferred_time_zone = self._output_time_zone
    configuration.profiling.directory = self._profiling_directory
    configuration.profiling.profilers = self._profilers
    configuration.profiling.sample_rate = self._profiling_sample_rate

    return configuration

  def _GetAnalysisPlugins(self, analysis_plugins_string):
    """Retrieves analysis plugins.

    Args:
      analysis_plugins_string (str): comma separated names of analysis plugins
          to enable.

    Returns:
      list[AnalysisPlugin]: analysis plugins.
    """
    if not analysis_plugins_string:
      return []

    analysis_plugins_list = [
        name.strip() for name in analysis_plugins_string.split(',')]

    analysis_plugins = self._analysis_manager.GetPluginObjects(
        analysis_plugins_list)
    return analysis_plugins.values()

  def _ParseAnalysisPluginOptions(self, options):
    """Parses the analysis plugin options.

    Args:
      options (argparse.Namespace): command line arguments.

    Raises:
      BadConfigOption: if the options are invalid.
    """
    # Get a list of all available plugins.
    analysis_plugin_info = self._analysis_manager.GetAllPluginInformation()
    # Use set-comprehension to create a set of the analysis plugin names.
    analysis_plugin_names = {
        name.lower() for name, _, _ in analysis_plugin_info}

    analysis_plugins = self.ParseStringOption(options, 'analysis_plugins')
    if not analysis_plugins:
      return

    # Use set-comprehension to create a set of the requested plugin names.
    requested_plugin_names = {
        name.strip().lower() for name in analysis_plugins.split(',')}

    # Check to see if we are trying to load plugins that do not exist.
    difference = requested_plugin_names.difference(analysis_plugin_names)
    if difference:
      difference_string = ' '.join(difference)
      raise errors.BadConfigOption(
          f'Non-existent analysis plugins specified: {difference_string:s}')

    self._analysis_plugins = self._GetAnalysisPlugins(analysis_plugins)

    for analysis_plugin in self._analysis_plugins:
      helpers_manager.ArgumentHelperManager.ParseOptions(
          options, analysis_plugin)

  def _ParseInformationalOptions(self, options):
    """Parses the informational options.

    Args:
      options (argparse.Namespace): command line arguments.

    Raises:
      BadConfigOption: if the options are invalid.
    """
    super()._ParseInformationalOptions(options)

    self._quiet_mode = getattr(options, 'quiet', False)

    helpers_manager.ArgumentHelperManager.ParseOptions(
        options, self, names=['status_view'])

  def _ParseProcessingOptions(self, options):
    """Parses the processing options.

    Args:
      options (argparse.Namespace): command line arguments.

    Raises:
      BadConfigOption: if the options are invalid.
    """
    argument_helper_names = [
        'process_resources', 'temporary_directory', 'zeromq']
    helpers_manager.ArgumentHelperManager.ParseOptions(
        options, self, names=argument_helper_names)

    worker_memory_limit = getattr(options, 'worker_memory_limit', None)

    if worker_memory_limit and worker_memory_limit < 0:
      raise errors.BadConfigOption((
          f'Invalid worker memory limit: {worker_memory_limit:d}, value must '
          f'be 0 or greater.'))

    worker_timeout = getattr(options, 'worker_timeout', None)

    if worker_timeout is not None and worker_timeout <= 0.0:
      raise errors.BadConfigOption((
          f'Invalid worker timeout: {worker_timeout:f}, value must be greater '
          f'than 0.0 minutes.'))

    self._worker_memory_limit = worker_memory_limit
    self._worker_timeout = worker_timeout

  def AddProcessingOptions(self, argument_group):
    """Adds processing options to the argument group.

    Args:
      argument_group (argparse._ArgumentGroup): argparse argument group.
    """
    argument_helper_names = ['temporary_directory', 'zeromq']
    if self._CanEnforceProcessMemoryLimit():
      argument_helper_names.append('process_resources')
    helpers_manager.ArgumentHelperManager.AddCommandLineArguments(
        argument_group, names=argument_helper_names)

    argument_group.add_argument(
        '--worker_memory_limit', '--worker-memory-limit',
        dest='worker_memory_limit', action='store', type=int,
        metavar='SIZE', help=(
            'Maximum amount of memory (data segment and shared memory) '
            'a worker process is allowed to consume in bytes, where 0 '
            'represents no limit. The default limit is 2147483648 (2 GiB). '
            'If a worker process exceeds this limit it is killed by the main '
            '(foreman) process.'))

    argument_group.add_argument(
        '--worker_timeout', '--worker-timeout', dest='worker_timeout',
        action='store', type=float, metavar='MINUTES', help=(
            'Number of minutes before a worker process that is not providing '
            'status updates is considered inactive. The default timeout is '
            '15.0 minutes. If a worker process exceeds this timeout it is '
            'killed by the main (foreman) process.'))

  def ListLanguageTags(self):
    """Lists the language tags."""
    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, column_names=['Language tag', 'Description'],
        title='Language tags')
    for language_tag, description in (
        language_tags.LanguageTagHelper.GetLanguages()):
      table_view.AddRow([language_tag, description])
    table_view.Write(self._output_writer)

  def ParseArguments(self, arguments):
    """Parses the command line arguments.

    Args:
      arguments (list[str]): command line arguments.

    Returns:
      bool: True if the arguments were successfully parsed.
    """
    loggers.ConfigureLogging()

    argument_parser = argparse.ArgumentParser(
        description=self.DESCRIPTION, add_help=False,
        conflict_handler='resolve',
        formatter_class=argparse.RawDescriptionHelpFormatter)

    self.AddBasicOptions(argument_parser)
    self.AddStorageOptions(argument_parser)

    analysis_group = argument_parser.add_argument_group('Analysis Arguments')

    helpers_manager.ArgumentHelperManager.AddCommandLineArguments(
        analysis_group, names=['analysis_plugins'])

    processing_group = argument_parser.add_argument_group('Processing')
    self.AddProcessingOptions(processing_group)

    info_group = argument_parser.add_argument_group('Informational Arguments')

    self.AddLogFileOptions(info_group)
    self.AddInformationalOptions(info_group)

    helpers_manager.ArgumentHelperManager.AddCommandLineArguments(
        info_group, names=['status_view'])

    filter_group = argument_parser.add_argument_group('Filter Arguments')

    helpers_manager.ArgumentHelperManager.AddCommandLineArguments(
        filter_group, names=['event_filters'])

    input_group = argument_parser.add_argument_group('Input Arguments')

    helpers_manager.ArgumentHelperManager.AddCommandLineArguments(
        input_group, names=['data_location'])

    output_group = argument_parser.add_argument_group('Output Arguments')

    output_group.add_argument(
        '-a', '--include_all', '--include-all', action='store_false',
        dest='dedup', default=True, help=(
            'By default the psort removes duplicate entries from the '
            'output. This parameter changes that behavior so all events '
            'are included.'))

    helpers_manager.ArgumentHelperManager.AddCommandLineArguments(
        output_group, names=['language'])

    self.AddOutputOptions(output_group)

    output_format_group = argument_parser.add_argument_group(
        'Output Format Arguments')

    helpers_manager.ArgumentHelperManager.AddCommandLineArguments(
        output_format_group, names=['output_modules'])

    profiling_group = argument_parser.add_argument_group('profiling arguments')

    helpers_manager.ArgumentHelperManager.AddCommandLineArguments(
        profiling_group, names=['profiling'])

    try:
      # TODO: refactor how arguments is used in a more argparse way.
      options = argument_parser.parse_args(arguments)
    except UnicodeEncodeError:
      # If we get here we are attempting to print help in a non-Unicode
      # terminal.
      self._output_writer.Write('\n')
      self._output_writer.Write(argument_parser.format_help())
      return False

    # Properly prepare the attributes according to local encoding.
    if self.preferred_encoding == 'ascii':
      self._PrintUserWarning((
          'the preferred encoding of your system is ASCII, which is not '
          'optimal for the typically non-ASCII characters that need to be '
          'parsed and processed. This will most likely result in an error.'))

    try:
      self.ParseOptions(options)
    except errors.BadConfigOption as exception:
      self._output_writer.Write(f'ERROR: {exception!s}\n')
      self._output_writer.Write('\n')
      self._output_writer.Write(argument_parser.format_usage())

      return False

    self._WaitUserWarning()

    loggers.ConfigureLogging(
        debug_output=self._debug_mode, filename=self._log_file,
        quiet_mode=self._quiet_mode)

    return True

  def ParseOptions(self, options):
    """Parses the options.

    Args:
      options (argparse.Namespace): command line arguments.

    Raises:
      BadConfigOption: if the options are invalid.
    """
    # The output modules options are dependent on the preferred_language
    # and output_time_zone options.
    self._ParseOutputOptions(options)

    names = ['analysis_plugins', 'language', 'profiling']
    helpers_manager.ArgumentHelperManager.ParseOptions(
        options, self, names=names)

    self.list_analysis_plugins = self._analysis_plugins == 'list'
    self.list_language_tags = self._preferred_language == 'list'
    self.list_profilers = self._profilers == 'list'

    self.show_troubleshooting = getattr(options, 'show_troubleshooting', False)

    if (self.list_analysis_plugins or self.list_language_tags or
        self.list_profilers or self.list_time_zones or
        self.show_troubleshooting):
      return

    # Check output modules after the other listable options, otherwise
    # it could raise with "requires an output file".
    helpers_manager.ArgumentHelperManager.ParseOptions(
        options, self, names=['output_modules'])

    self.list_output_modules = self._output_format == 'list'
    if self.list_output_modules:
      return

    self._ParseInformationalOptions(options)

    helpers_manager.ArgumentHelperManager.ParseOptions(
        options, self, names=['data_location'])

    self._ParseLogFileOptions(options)

    self._ParseProcessingOptions(options)

    helpers_manager.ArgumentHelperManager.ParseOptions(
        options, self, names=['event_filters'])

    self._deduplicate_events = getattr(options, 'dedup', True)

    if self._data_location:
      # Update the data location with the calculated value.
      options.data_location = self._data_location
    else:
      logger.warning('Unable to automatically determine data location.')

    self._command_line_arguments = self.GetCommandLineArguments()

    self._storage_file_path = self.ParseStringOption(options, 'storage_file')

    self._EnforceProcessMemoryLimit(self._process_memory_limit)

    self._analysis_plugins = self._CreateAnalysisPlugins(options)
    self._output_module = self._CreateOutputModule(options)

    check_readable_only = not self._analysis_plugins
    self._CheckStorageFile(
        self._storage_file_path, check_readable_only=check_readable_only)

  def ProcessStorage(self):
    """Processes a Plaso storage file.

    Raises:
      BadConfigOption: when a configuration parameter fails validation or the
          storage file cannot be opened with read access.
      RuntimeError: if a non-recoverable situation is encountered.
    """
    self._status_view.SetMode(self._status_view_mode)
    self._status_view.SetStatusFile(self._status_view_file)
    self._status_view.SetStorageFileInformation(self._storage_file_path)

    status_update_callback = (
        self._status_view.GetAnalysisStatusUpdateCallback())

    storage_reader = storage_factory.StorageFactory.CreateStorageReaderForFile(
        self._storage_file_path)
    if not storage_reader:
      raise RuntimeError('Unable to create storage reader.')

    try:
      self._number_of_stored_analysis_reports = (
          storage_reader.GetNumberOfAttributeContainers(
              self._CONTAINER_TYPE_ANALYSIS_REPORT))
    finally:
      storage_reader.Close()

    session = engine.BaseEngine.CreateSession()

    configuration = self._CreateOutputAndFormattingProcessingConfiguration()

    # TODO: implement _CreateAnalysisProcessingConfiguration

    if self._analysis_plugins:
      self._AnalyzeEvents(
          session, configuration, status_update_callback=status_update_callback)

    # TODO: abort if session.aborted is True

    if self._output_format != 'null':
      storage_reader = (
          storage_factory.StorageFactory.CreateStorageReaderForFile(
              self._storage_file_path))

      # TODO: add single process output and formatting engine support.
      output_engine = (
          multi_output_engine.OutputAndFormattingMultiProcessEngine())

      output_engine.SetStatusUpdateInterval(self._status_view_interval)

      output_engine.ExportEvents(
          storage_reader, self._output_module, configuration,
          deduplicate_events=self._deduplicate_events,
          event_filter=self._event_filter,
          status_update_callback=status_update_callback,
          time_slice=self._time_slice, use_time_slicer=self._use_time_slicer)

      self._output_module.Close()
      self._output_module = None

    if self._quiet_mode:
      return

    self._output_writer.Write('Processing completed.\n')

    storage_reader = storage_factory.StorageFactory.CreateStorageReaderForFile(
        self._storage_file_path)
    self._PrintAnalysisReportsDetails(storage_reader)



"""Parser for Safari Binary Cookie files."""

import os

from dfdatetime import cocoa_time as dfdatetime_cocoa_time

from dtfabric.runtime import data_maps as dtfabric_data_maps

from plaso.containers import events
from plaso.lib import cookie_plugins_helper
from plaso.lib import dtfabric_helper
from plaso.lib import errors
from plaso.lib import specification
from plaso.parsers import interface
from plaso.parsers import manager


class SafariBinaryCookieEventData(events.EventData):
  """Safari binary cookie event data.

  Attributes:
    cookie_name (str): cookie name.
    cookie_value (str): cookie value.
    creation_time (dfdatetime.DateTimeValues): date and time the cookie
        was created.
    expiration_time (dfdatetime.DateTimeValues): date and time the cookie
        expires.
    flags (int): cookie flags.
    path (str): path of the cookie.
    url (str): URL where this cookie is valid.
  """

  DATA_TYPE = 'safari:cookie:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.cookie_name = None
    self.cookie_value = None
    self.creation_time = None
    self.expiration_time = None
    self.flags = None
    self.path = None
    self.url = None


class BinaryCookieParser(
    interface.FileObjectParser, cookie_plugins_helper.CookiePluginsHelper,
    dtfabric_helper.DtFabricHelper):
  """Parser for Safari Binary Cookie files."""

  NAME = 'binary_cookies'
  DATA_FORMAT = 'Safari Binary Cookie file'

  _DEFINITION_FILE = os.path.join(
      os.path.dirname(__file__), 'safari_cookies.yaml')

  def _ParseCString(self, page_data, string_offset):
    """Parses a C string from the page data.

    Args:
      page_data (bytes): page data.
      string_offset (int): offset of the string relative to the start
          of the page.

    Returns:
      str: string.

    Raises:
      ParseError: when the string cannot be parsed.
    """
    cstring_map = self._GetDataTypeMap('cstring')

    try:
      value_string = self._ReadStructureFromByteStream(
          page_data[string_offset:], string_offset, cstring_map)
    except (ValueError, errors.ParseError) as exception:
      raise errors.ParseError((
          f'Unable to map string data at offset: 0x{string_offset:08x} '
          f'with error: {exception!s}'))

    return value_string.rstrip('\x00')

  def _ParsePage(self, parser_mediator, file_offset, page_data):
    """Parses a page.

    Args:
      parser_mediator (ParserMediator): parser mediator.
      file_offset (int): offset of the data relative from the start of
          the file-like object.
      page_data (bytes): page data.

    Raises:
      ParseError: when the page cannot be parsed.
    """
    page_header_map = self._GetDataTypeMap('binarycookies_page_header')

    try:
      page_header = self._ReadStructureFromByteStream(
          page_data, file_offset, page_header_map)
    except (ValueError, errors.ParseError) as exception:
      raise errors.ParseError((
          f'Unable to map page header data at offset: 0x{file_offset:08x} '
          f'with error: {exception!s}'))

    for record_offset in page_header.offsets:
      if parser_mediator.abort:
        break

      self._ParseRecord(parser_mediator, page_data, record_offset)

  def _ParseRecord(self, parser_mediator, page_data, record_offset):
    """Parses a record from the page data.

    Args:
      parser_mediator (ParserMediator): parser mediator.
      page_data (bytes): page data.
      record_offset (int): offset of the record relative to the start
          of the page.

    Raises:
      ParseError: when the record cannot be parsed.
    """
    record_header_map = self._GetDataTypeMap('binarycookies_record_header')

    try:
      record_header = self._ReadStructureFromByteStream(
          page_data[record_offset:], record_offset, record_header_map)
    except (ValueError, errors.ParseError) as exception:
      raise errors.ParseError((
          f'Unable to map record header data at offset: 0x{record_offset:08x} '
          f'with error: {exception!s}'))

    event_data = SafariBinaryCookieEventData()
    event_data.flags = record_header.flags

    if record_header.url_offset:
      data_offset = record_offset + record_header.url_offset
      event_data.url = self._ParseCString(page_data, data_offset)

    if record_header.name_offset:
      data_offset = record_offset + record_header.name_offset
      event_data.cookie_name = self._ParseCString(page_data, data_offset)

    if record_header.path_offset:
      data_offset = record_offset + record_header.path_offset
      event_data.path = self._ParseCString(page_data, data_offset)

    if record_header.value_offset:
      data_offset = record_offset + record_header.value_offset
      event_data.cookie_value = self._ParseCString(page_data, data_offset)

    if record_header.creation_time:
      event_data.creation_time = dfdatetime_cocoa_time.CocoaTime(
          timestamp=record_header.creation_time)

    if record_header.expiration_time:
      event_data.expiration_time = dfdatetime_cocoa_time.CocoaTime(
          timestamp=record_header.expiration_time)

    parser_mediator.ProduceEventData(event_data)

    self._ParseCookie(
        parser_mediator, event_data.cookie_name, event_data.cookie_value,
        event_data.cookie_name)

  @classmethod
  def GetFormatSpecification(cls):
    """Retrieves the format specification for parser selection.

    Returns:
      FormatSpecification: format specification.
    """
    format_specification = specification.FormatSpecification(cls.NAME)
    format_specification.AddNewSignature(b'cook\x00', offset=0)
    return format_specification

  def ParseFileObject(self, parser_mediator, file_object):
    """Parses a Safari binary cookie file-like object.

    Args:
      parser_mediator (ParserMediator): parser mediator.
      file_object (dfvfs.FileIO): file-like object to be parsed.

    Raises:
      ParseError: when the page sizes array cannot be parsed.
      WrongParser: when the file cannot be parsed, this will signal
          the event extractor to apply other parsers.
    """
    file_header_map = self._GetDataTypeMap('binarycookies_file_header')

    try:
      file_header, file_header_data_size = self._ReadStructureFromFileObject(
          file_object, 0, file_header_map)
    except (ValueError, errors.ParseError) as exception:
      raise errors.WrongParser(
          f'Unable to read file header with error: {exception!s}.')

    file_offset = file_header_data_size

    # TODO: move page sizes array into file header, this will require dtFabric
    # to compare signature as part of data map.
    page_sizes_data_size = file_header.number_of_pages * 4

    page_sizes_data = file_object.read(page_sizes_data_size)

    context = dtfabric_data_maps.DataTypeMapContext(values={
        'binarycookies_file_header': file_header})

    page_sizes_map = self._GetDataTypeMap('binarycookies_page_sizes')

    try:
      page_sizes_array = self._ReadStructureFromByteStream(
          page_sizes_data, file_offset, page_sizes_map, context=context)
    except (ValueError, errors.ParseError) as exception:
      raise errors.ParseError((
          f'Unable to map page sizes data at offset: 0x{file_offset:08x} '
          f'with error: {exception!s}'))

    file_offset += page_sizes_data_size

    for page_number, page_size in enumerate(page_sizes_array):
      if parser_mediator.abort:
        break

      page_data = file_object.read(page_size)
      if len(page_data) != page_size:
        parser_mediator.ProduceExtractionWarning(
            f'unable to read page: {page_number:d}')
        break

      self._ParsePage(parser_mediator, file_offset, page_data)

      file_offset += page_size

    # TODO: check file footer.


manager.ParsersManager.RegisterParser(BinaryCookieParser)



"""Text parser plugin for WinCC log files."""

from dfvfs.helpers import text_file

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.lib import errors
from plaso.parsers import interface
from plaso.parsers import manager


class SIMATICS7EventData(events.EventData):
  """SIMATIC S7 event data.

  Attributes:
    body (str): the message content of the event.
    creation_time (dfdatetime.DateTimeValues): date and time the log entry
        was created.
  """

  DATA_TYPE = 'wincc:simatic_s7:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.body = None
    self.creation_time = None


class WinCCSysLogEventData(events.EventData):
  """WinCC Sys Log event data.

  Attributes:
    body (str): the content of the log's message.
    creation_time (dfdatetime.DateTimeValues): date and time the log entry
        was created.
    event_number (int): a number specifying the type of event.
    log_hostname (str): the hostname of the machine logging the event.
    log_identifier (int): identifier for this log file.
    source_device (str): which device generated the event.
  """

  DATA_TYPE = 'wincc:sys_log:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.body = None
    self.creation_time = None
    self.event_number = None
    self.log_hostname = None
    self.log_identifier = None
    self.source_device = None


class SIMATICLogParser(interface.FileObjectParser):
  """Text parser plugin for SIMATIC S7 Log files."""

  NAME = 'simatic_s7'
  DATA_FORMAT = 'SIMATIC S7 Log file'

  DELIMITER = ','
  ENCODING = 'ascii'
  END_OF_LINE = '\r\n'

  _EXPECTED_FIRST_LINE_STRING = 'Log starting ...' + END_OF_LINE
  _EXPECTED_SECOND_LINE_STRING = '| LogFileName'
  _EXPECTED_THIRD_LINE_STRING = '| LogFileCount'

  def _ParseValues(self, parser_mediator, line_number, values):
    """Parses SIMATIC S7 log file values.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      line_number (int): number of the line the values were extracted from.
      values (list[str]): values extracted from the line.

    Raises:
      WrongParser: when the values cannot be parsed.
    """
    number_of_values = len(values)
    if number_of_values < 2:
      raise errors.WrongParser(
          f'Expected at least two values on line {line_number:d}')

    # The first lines seem to  always follow the following format:
    #
    # 2019-05-27 10:05:43,405 INFO     Log starting ...
    # 2019-05-27 10:05:43,405 INFO     | LogFileName   : C:\.....
    # 2019-05-27 10:05:43,419 INFO     | LogFileCount  : 3
    if line_number == 0:
      if not values[1].endswith(self._EXPECTED_FIRST_LINE_STRING):
        raise errors.WrongParser(
            f'Expected first line to end with: '
            f'"{self._EXPECTED_FIRST_LINE_STRING:s}"')

    if line_number == 1:
      if values[1].find(self._EXPECTED_SECOND_LINE_STRING) < 0:
        raise errors.WrongParser(
            f'Expected second line to contain: '
            f'"{self._EXPECTED_SECOND_LINE_STRING:s}"')

    if line_number == 2:
      if values[1].find(self._EXPECTED_THIRD_LINE_STRING) < 0:
        raise errors.WrongParser(
            f'Expected third line to contain: '
            f'"{self._EXPECTED_THIRD_LINE_STRING:s}"')

    event_data = SIMATICS7EventData()
    try:
      date_time = dfdatetime_time_elements.TimeElements()
      date_time.CopyFromDateTimeString(values[0])
      event_data.creation_time = date_time
    except (TypeError, ValueError) as exception:
      error_string = (
          f'Unable to parse date time value with error: '
          f'{exception!s} on line {line_number:d} {values!s}')
      parser_mediator.ProduceExtractionWarning(error_string)

    event_data.body = ','.join(values[1:]).strip()

    parser_mediator.ProduceEventData(event_data)

  def ParseFileObject(self, parser_mediator, file_object):
    """Parses a SIMATIC Log file-like object.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      file_object (dfvfs.FileIO): file-like object.

    Raises:
      WrongParser: when the file cannot be parsed.
    """
    line_reader = text_file.TextFile(
        file_object, encoding=self.ENCODING, end_of_line=self.END_OF_LINE)

    line_number = 0

    try:
      line = line_reader.readline()
    except UnicodeDecodeError as exception:
      raise errors.WrongParser(
          f'unable to read line: {line_number:d} with error: {exception!s}')

    while line:
      values = line.split(self.DELIMITER)
      self._ParseValues(parser_mediator, line_number, values)

      try:
        line_number += 1
        line = line_reader.readline()
      except UnicodeDecodeError as exception:
        parser_mediator.ProduceExtractionWarning(
            f'unable to read line: {line_number:d} with error: {exception!s}')
        break


class WinCCSysLogParser(interface.FileObjectParser):
  """Text parser plugin for WinCC Sys Log files."""

  NAME = 'wincc_sys'
  DATA_FORMAT = 'WinCC Sys Log file'

  DELIMITER = ','
  ENCODING = 'utf-16-le'

  _DISALLOWED_HOSTNAME_CHARS = ['\\', '/', ':', '*', '?', '"', '<', '>', '|']

  _END_OF_LOG_FILE_STRING = '======>'

  def _ParseValues(self, parser_mediator, line_number, values, first_line):
    """Parses WinCC log file values.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      line_number (int): number of the line the values were extracted from.
      values (list[str]): values extracted from the line.
      first_line (bool): True if this is first line from which values were
          extracted.

    Raises:
      WrongParser: when the values cannot be parsed.
    """
    number_of_values = len(values)
    if len(values) < 10:
      error_string = (
          f'invalid number of values: {number_of_values:d} in line: '
          f'{line_number:d}')

      # On other lines, we might encounter times where the split() operation
      # produces more values.
      if first_line:
        raise errors.WrongParser(error_string)

      parser_mediator.ProduceExtractionWarning(error_string)

    event_data = WinCCSysLogEventData()

    identifier_value = values[0]

    try:
      log_identifier = int(identifier_value, 10)
      event_data.log_identifier = log_identifier
    except ValueError as exception:
      self._ParseValuesFail(parser_mediator, first_line, exception, (
          f'Type of first value ({identifier_value!s}) should be an int in '
          f'line: {line_number:d}'))

    date_value = values[1]
    time_value = values[2]

    try:
      day_of_month, month, year = [
          int(element, 10) for element in date_value.split('.')]
      hours, minutes, seconds, milliseconds = [
          int(element, 10) for element in time_value.split(':')]
      time_elements_tuple = (
          year, month, day_of_month, hours, minutes, seconds, milliseconds)
      date_time = dfdatetime_time_elements.TimeElementsInMilliseconds(
          time_elements_tuple=time_elements_tuple)

      # TODO: determine if format is dependant on the system's locale.
      date_time.is_local_time = True

      event_data.creation_time = date_time

    except (TypeError, ValueError) as exception:
      self._ParseValuesFail(parser_mediator, first_line, exception, (
          f'Unable to parse time elements with error: {exception!s} on line '
          f'{line_number:d} [{date_value!s}, {time_value!s}]'))

    event_number_value = values[3]

    try:
      event_data.event_number = int(event_number_value, 10)
    except ValueError as exception:
      self._ParseValuesFail(parser_mediator, first_line, exception, (
          f'Type of event_number value ({event_number_value!s}) should be a '
          f'decimal in line: {line_number:d}'))

    # Column 4 and 5 contain unknown values, we are not parsing them.

    # We are checking this only once, on the first line, as we expect all
    # following lines to contain the same hostname, as the logs are
    # collected from the system generating it.
    # Using this documentation to validate a Windows host name.
    # https://learn.microsoft.com/en-us/troubleshoot/windows-server/identity/naming-conventions-for-computer-domain-site-ou
    hostname = values[6]

    if first_line:
      if not hostname or len(hostname) > 16:
        self._ParseValuesFail(parser_mediator, first_line, exception, (
            f'Unsupported hostname: "{hostname!s}" on line: {line_number:d}'))

      for character in self._DISALLOWED_HOSTNAME_CHARS:
        if character in hostname:
          self._ParseValuesFail(parser_mediator, first_line, exception, (
              f'Unsupported hostname: "{hostname!s}" on line: {line_number:d}'))

    event_data.log_hostname = hostname

    event_data.source_device = values[7]

    # The second to last field is the one that might contain unquoted separator.
    # The last field can contain a string such as MSG_STATE_COME, MSG_STATE_GO,
    # but also sometimes contains a message string.

    text_message = ' '.join(values[8:])
    event_data.body = text_message

    parser_mediator.ProduceEventData(event_data)

  def _ParseValuesFail(
      self, parser_mediator, first_line, parent_exception, error_string):
    if first_line:
      raise errors.WrongParser(error_string) from parent_exception

    parser_mediator.ProduceExtractionWarning(error_string)

  def ParseFileObject(self, parser_mediator, file_object):
    """Parses a WinCC Sys Log file-like object.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      file_object (dfvfs.FileIO): file-like object.

    Raises:
      WrongParser: when the file cannot be parsed.
    """
    # Note that we cannot use the DSVParser here since this WinCC file format is
    # not strict and clean file format.
    line_reader = text_file.TextFile(
        file_object, encoding=self.ENCODING, end_of_line='\r\n')

    line_number = 0
    first_line = True

    # While the file is in UTF-16, the end_of_line character is '\r\n'.
    # We manually strip this out.

    try:
      line = line_reader.readline().strip()
    except UnicodeDecodeError as exception:
      raise errors.WrongParser(
          f'unable to read line: {line_number:d} with error: {exception!s}')

    while line:
      if line.startswith(self._END_OF_LOG_FILE_STRING):
        # It seems that sometimes WinCC logs will end with a line starting like
        # this, with the name of the next log file.
        # But this is not always the case though.
        line = line_reader.readline().strip()
        continue
      values = line.split(self.DELIMITER)
      self._ParseValues(parser_mediator, line_number, values, first_line)

      try:
        line_number += 1
        first_line = False
        line = line_reader.readline().strip()
      except UnicodeDecodeError as exception:
        parser_mediator.ProduceExtractionWarning(
            f'unable to read line: {line_number:d} with error: {exception!s}')
        break


manager.ParsersManager.RegisterParsers([WinCCSysLogParser, SIMATICLogParser])



"""Parser for locate database (updatedb) files."""

import os

from dfdatetime import posix_time

from plaso.containers import events
from plaso.lib import definitions
from plaso.lib import dtfabric_helper
from plaso.lib import errors
from plaso.lib import specification
from plaso.parsers import interface
from plaso.parsers import manager


class LocateDatabaseEvent(events.EventData):
  """Linux locate database (updatedb) event data.

  Attributes:
    entries (list[str]): contents of the locate database (updatedb) entry.
    path (str): path of the locate database (updatedb) entry.
    written_time (dfdatetime.DateTimeValues): entry written date and time.
  """

  DATA_TYPE = 'linux:locate_database:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.entries = None
    self.path = None
    self.written_time = None


class LocateDatabaseParser(
    interface.FileObjectParser, dtfabric_helper.DtFabricHelper):
  """Parser for locate database (updatedb) files."""

  NAME = 'locate_database'
  DATA_FORMAT = 'Locate database file (updatedb)'

  _DEFINITION_FILE = os.path.join(
      os.path.dirname(__file__), 'locate.yaml')

  def __init__(self):
    """Initializes a locate database (updatedb) file parser."""
    super().__init__()
    self._cstring_map = self._GetDataTypeMap('cstring')
    self._directory_entry_header_map = self._GetDataTypeMap(
        'directory_entry_header')
    self._directory_header_map = self._GetDataTypeMap('directory_header')

  def _ParseDirectoryEntry(self, file_object, file_offset):
    """Parses a locate database (updatedb) directory entry.

    Args:
      file_object (dfvfs.FileIO): file-like object to be parsed.
      file_offset (int): offset of the directory entry relative to the start of
          the file.

    Returns:
      tuple[list[str], int]: names of sub directory entries and total number of
          bytes read.
    """
    sub_entry_names = []
    total_data_size = 0

    # TODO: determine why "condition: directory_entry.type != 2" in dtFabric
    # definitions is currently not working and clean up code once fixed.
    directory_entry_type = 0
    while directory_entry_type != 2:
      directory_entry_header, data_size = self._ReadStructureFromFileObject(
          file_object, file_offset, self._directory_entry_header_map)

      file_offset += data_size
      total_data_size += data_size

      directory_entry_type = directory_entry_header.type
      if directory_entry_type != 2:
        directory_entry_path, data_size = self._ReadStructureFromFileObject(
            file_object, file_offset, self._cstring_map)

        sub_entry_names.append(directory_entry_path)

        file_offset += data_size
        total_data_size += data_size

    return sub_entry_names, total_data_size

  @classmethod
  def GetFormatSpecification(cls):
    """Retrieves the format specification.

    Returns:
      FormatSpecification: format specification.
    """
    format_specification = specification.FormatSpecification(cls.NAME)
    format_specification.AddNewSignature(b'\x00mlocate', offset=0)
    return format_specification

  # pylint: disable=unused-argument
  def ParseFileObject(self, parser_mediator, file_object, **kwargs):
    """Parses a locate database (updatedb) file-like object.

    Args:
      parser_mediator (ParserMediator): parser mediator.
      file_object (dfvfs.FileIO): file-like object to be parsed.

    Raises:
      WrongParser: when the file cannot be parsed, this will signal
          the event extractor to apply other parsers.
    """
    locate_database_header_map = self._GetDataTypeMap('locate_database_header')

    try:
      locate_database_header, file_offset = self._ReadStructureFromFileObject(
          file_object, 0, locate_database_header_map)
    except (ValueError, errors.ParseError) as exception:
      raise errors.WrongParser(
          f'Unable to parse locate database header with error: {exception!s}')

    # Skip configuration block for now.
    file_offset += locate_database_header.configuration_block_size

    file_size = file_object.get_size()
    while file_offset + 16 < file_size:
      try:
        directory_header, data_size = self._ReadStructureFromFileObject(
            file_object, file_offset, self._directory_header_map)
      except (ValueError, errors.ParseError) as exception:
        parser_mediator.ProduceExtractionWarning(
            f'unable to parse locate directory header at offset: '
            f'0x{file_offset:08x} with error: {exception!s}')
        return

      file_offset += data_size

      timestamp = directory_header.nanoseconds + (
          directory_header.seconds * definitions.NANOSECONDS_PER_SECOND)

      try:
        entries, data_size = self._ParseDirectoryEntry(
            file_object, file_offset)

        file_offset += data_size

      except (ValueError, errors.ParseError) as exception:
        parser_mediator.ProduceExtractionWarning(
            f'unable to parse directory entry at offset: '
            f'0x{file_offset:08x} with error: {exception!s}')
        return

      event_data = LocateDatabaseEvent()
      event_data.entries = entries or None
      event_data.path = directory_header.path
      event_data.written_time = posix_time.PosixTimeInNanoseconds(
          timestamp=timestamp)

      parser_mediator.ProduceEventData(event_data)


manager.ParsersManager.RegisterParser(LocateDatabaseParser)



"""SQLite parser plugin for iOS Screen Time database files."""

from dfdatetime import cocoa_time as dfdatetime_cocoa_time

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class IOSScreenTimeEventData(events.EventData):
  """iOS Screen Time file usage event data.

  Attributes:
    bundle_identifier (str): Bundle Identifier of the application.
    device_identifier (str): GUID for the device.
    device_name (str): Name of the device in use (when available).
    domain (str): Domain of the website visited.
    start_time (dfdatetime.DateTimeValues): date and time the start of
        the application.
    total_time (int): Number of seconds where the application was in the
        foreground.
    user_family_name (str): Family name of the user.
    user_given_name (str): Given name of the user.
  """

  DATA_TYPE = 'ios:screentime:event'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.bundle_identifier = None
    self.device_identifier = None
    self.device_name = None
    self.domain = None
    self.start_time = None
    self.total_time = None
    self.user_family_name = None
    self.user_given_name = None


class IOSScreenTimePlugin(interface.SQLitePlugin):
  """SQLite parser plugin for iOS Screen Time database files.

  The Screen Time database is typically stored in:
  RMAdminStore-Local.sqlite
  """

  NAME = 'ios_screentime'
  DATA_FORMAT = 'iOS Screen Time SQLite database (RMAdminStore-Local.sqlite)'

  REQUIRED_STRUCTURE = {
      'ZUSAGETIMEDITEM': frozenset([
          'ZBUNDLEIDENTIFIER', 'ZCATEGORY', 'ZDOMAIN', 'ZTOTALTIMEINSECONDS']),
      'ZUSAGECATEGORY': frozenset(['Z_PK', 'ZBLOCK']),
      'ZUSAGEBLOCK': frozenset(['Z_PK', 'ZSTARTDATE', 'ZUSAGE']),
      'ZUSAGE': frozenset(['Z_PK', 'ZDEVICE', 'ZUSER']),
      'ZCOREDEVICE': frozenset(['Z_PK', 'ZIDENTIFIER', 'ZNAME']),
      'ZCOREUSER': frozenset(['Z_PK', 'ZFAMILYNAME', 'ZGIVENNAME'])}

  QUERIES = [(
      """
      SELECT ZUSAGETIMEDITEM.ZTOTALTIMEINSECONDS,
        ZUSAGETIMEDITEM.ZBUNDLEIDENTIFIER,
        ZUSAGETIMEDITEM.ZDOMAIN,
        ZUSAGEBLOCK.ZSTARTDATE,
        ZCOREDEVICE.ZIDENTIFIER,
        ZCOREDEVICE.ZNAME,
        ZCOREUSER.ZFAMILYNAME,
        ZCOREUSER.ZGIVENNAME
      FROM ZUSAGETIMEDITEM
      LEFT JOIN ZUSAGECATEGORY
        ON ZUSAGETIMEDITEM.ZCATEGORY = ZUSAGECATEGORY.Z_PK
      LEFT JOIN ZUSAGEBLOCK ON ZUSAGECATEGORY.ZBLOCK = ZUSAGEBLOCK.Z_PK
      LEFT JOIN ZUSAGE ON ZUSAGEBLOCK.ZUSAGE = ZUSAGE.Z_PK
      LEFT JOIN ZCOREDEVICE ON ZUSAGE.ZDEVICE = ZCOREDEVICE.Z_PK
      LEFT JOIN ZCOREUSER ON ZUSAGE.ZUSER = ZCOREUSER.Z_PK
      """, 'ParseScreenTimeRow')]

  SCHEMAS = {
      'ZUSAGETIMEDITEM': (
          'CREATE TABLE ZUSAGETIMEDITEM ( Z_PK INTEGER PRIMARY KEY, '
          'Z_ENT INTEGER, Z_OPT INTEGER, ZTOTALTIMEINSECONDS INTEGER, '
          'ZUSAGETRUSTED INTEGER, ZCATEGORY INTEGER, ZBUNDLEIDENTIFIER VARCHAR,'
          ' ZDOMAIN VARCHAR )'),
      'ZUSAGECATEGORY': (
          'CREATE TABLE ZUSAGECATEGORY ( Z_PK INTEGER PRIMARY KEY, '
          'Z_ENT INTEGER, Z_OPT INTEGER, ZTOTALTIMEINSECONDS INTEGER, '
          'ZBLOCK INTEGER, ZIDENTIFIER VARCHAR )'),
      'ZUSAGEBLOCK': (
          'CREATE TABLE ZUSAGEBLOCK ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
          'Z_OPT INTEGER, ZDURATIONINMINUTES INTEGER, '
          'ZNUMBEROFPICKUPSWITHOUTAPPLICATIONUSAGE INTEGER, '
          'ZSCREENTIMEINSECONDS INTEGER, ZUSAGE INTEGER, '
          'ZFIRSTPICKUPDATE TIMESTAMP, ZLASTEVENTDATE TIMESTAMP, '
          'ZLONGESTSESSIONENDDATE TIMESTAMP, '
          'ZLONGESTSESSIONSTARTDATE TIMESTAMP, ZSTARTDATE TIMESTAMP )'),
      'ZUSAGE': (
          'CREATE TABLE ZUSAGE ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
          'Z_OPT INTEGER, ZDEVICE INTEGER, ZUSER INTEGER, '
          'ZLASTEVENTDATE TIMESTAMP, ZLASTUPDATEDDATE TIMESTAMP )'),
      'ZCOREDEVICE': (
          'CREATE TABLE ZCOREDEVICE ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
          'Z_OPT INTEGER, ZPLATFORM INTEGER, ZLOCALUSERDEVICESTATE INTEGER, '
          'ZIDENTIFIER VARCHAR, ZNAME VARCHAR )'),
      'ZCOREUSER': (
          'CREATE TABLE ZCOREUSER ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
          'Z_OPT INTEGER, ZDSID INTEGER, ZISFAMILYORGANIZER INTEGER, '
          'ZISPARENT INTEGER, ZPASSCODEENTRYATTEMPTCOUNT INTEGER, '
          'ZPASSCODERECOVERYATTEMPTCOUNT INTEGER, ZSUPPORTSENCRYPTION INTEGER, '
          'ZCLOUDSETTINGS INTEGER, ZFAMILYSETTINGS INTEGER, '
          'ZLOCALSETTINGS INTEGER, ZLOCALUSERDEVICESTATE INTEGER, '
          'ZPASSCODEENTRYTIMEOUTENDDATE TIMESTAMP, ZALTDSID VARCHAR, '
          'ZAPPLEID VARCHAR, ZFAMILYMEMBERTYPE VARCHAR, ZFAMILYNAME VARCHAR, '
          'ZGIVENNAME VARCHAR, ZPHONETICFAMILYNAME VARCHAR, '
          'ZPHONETICGIVENNAME VARCHAR )')}

  REQUIRES_SCHEMA_MATCH = False

  def _GetDateTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.CocoaTime: date and time value or None if not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    if timestamp is None:
      return None

    return dfdatetime_cocoa_time.CocoaTime(timestamp=timestamp)

  # pylint: disable=unused-argument
  def ParseScreenTimeRow(self, parser_mediator, query, row, **unused_kwargs):
    """Parses a Screen Time row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    event_data = IOSScreenTimeEventData()
    event_data.bundle_identifier = self._GetRowValue(
        query_hash, row, 'ZBUNDLEIDENTIFIER')
    event_data.device_identifier = self._GetRowValue(
        query_hash, row, 'ZIDENTIFIER')
    event_data.device_name = self._GetRowValue(query_hash, row, 'ZNAME')
    event_data.domain = self._GetRowValue(query_hash, row, 'ZDOMAIN')
    event_data.start_time = self._GetDateTimeRowValue(
        query_hash, row, 'ZSTARTDATE')
    event_data.total_time = self._GetRowValue(
        query_hash, row, 'ZTOTALTIMEINSECONDS')
    event_data.user_family_name = self._GetRowValue(
        query_hash, row, 'ZFAMILYNAME')
    event_data.user_given_name = self._GetRowValue(
        query_hash, row, 'ZGIVENNAME')

    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(IOSScreenTimePlugin)



"""Operating system independent (generic) preprocessor plugins."""

from dfvfs.helpers import file_system_searcher

from plaso.lib import definitions
from plaso.preprocessors import interface
from plaso.preprocessors import manager


class DetermineOperatingSystemPlugin(
    interface.FileSystemArtifactPreprocessorPlugin):
  """Plugin to determine the operating system."""

  # pylint: disable=abstract-method
  # This plugin does not use an artifact definition and therefore does not
  # use _ParsePathSpecification.

  # We need to check for both forward and backward slashes since the path
  # specification will be dfVFS back-end dependent.
  _WINDOWS_LOCATIONS = set([
      '/windows/system32', '\\windows\\system32', '/winnt/system32',
      '\\winnt\\system32', '/winnt35/system32', '\\winnt35\\system32',
      '\\wtsrv\\system32', '/wtsrv/system32'])

  def __init__(self):
    """Initializes a plugin to determine the operating system."""
    super().__init__()
    self._find_specs = [
        file_system_searcher.FindSpec(
            case_sensitive=False, location='/etc',
            location_separator='/'),
        file_system_searcher.FindSpec(
            case_sensitive=False, location='/System/Library',
            location_separator='/'),
        file_system_searcher.FindSpec(
            case_sensitive=False, location='\\Windows\\System32',
            location_separator='\\'),
        file_system_searcher.FindSpec(
            case_sensitive=False, location='\\WINNT\\System32',
            location_separator='\\'),
        file_system_searcher.FindSpec(
            case_sensitive=False, location='\\WINNT35\\System32',
            location_separator='\\'),
        file_system_searcher.FindSpec(
            case_sensitive=False, location='\\WTSRV\\System32',
            location_separator='\\')]

  # pylint: disable=unused-argument
  def Collect(self, mediator, artifact_definition, searcher, file_system):
    """Collects values using a file artifact definition.

    Args:
      mediator (PreprocessMediator): mediates interactions between preprocess
          plugins and other components, such as storage and knowledge base.
      artifact_definition (artifacts.ArtifactDefinition): artifact definition.
      searcher (dfvfs.FileSystemSearcher): file system searcher to preprocess
          the file system.
      file_system (dfvfs.FileSystem): file system to be preprocessed.

    Raises:
      PreProcessFail: if the preprocessing fails.
    """
    locations = []
    for path_spec in searcher.Find(find_specs=self._find_specs):
      relative_path = searcher.GetRelativePath(path_spec)
      if relative_path:
        locations.append(relative_path.lower())

    operating_system = definitions.OPERATING_SYSTEM_FAMILY_UNKNOWN
    if self._WINDOWS_LOCATIONS.intersection(set(locations)):
      operating_system = definitions.OPERATING_SYSTEM_FAMILY_WINDOWS_NT

    elif '/system/library' in locations:
      operating_system = definitions.OPERATING_SYSTEM_FAMILY_MACOS

    elif '/etc' in locations:
      operating_system = definitions.OPERATING_SYSTEM_FAMILY_LINUX

    if operating_system != definitions.OPERATING_SYSTEM_FAMILY_UNKNOWN:
      mediator.SetValue('operating_system', operating_system)


manager.PreprocessPluginsManager.RegisterPlugins([
    DetermineOperatingSystemPlugin])



"""SQLite parser plugin for Android Burner database files."""

from dfdatetime import java_time as dfdatetime_java_time

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class AndroidBurnerEventData(events.EventData):
  """Android Burner event data.

  Attributes:
    alias (str): Alias or nickname.
    creation_time (dfdatetime.DateTimeValues): Creation date and time.
    expiration_time (dfdatetime.DateTimeValues): Expiration date and time.
    last_updated_time (dfdatetime.DateTimeValues): Last update date and time.
    name (str): Name of the burner.
    phone_number (str): Associated phone number.
    total_minutes (int): Total minutes available for calls.
    user_identifier (str): User identifier.
    voicemail_url (str): URL for accessing the voicemail associated with the
        burner.
  """

  DATA_TYPE = 'android:burners'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.alias = None
    self.creation_time = None
    self.expiration_time = None
    self.last_updated_time = None
    self.name = None
    self.phone_number = None
    self.total_minutes = None
    self.user_identifier = None
    self.voicemail_url = None


class AndroidBurnerPlugin(interface.SQLitePlugin):
  """SQLite parser plugin for Android communication information database files.

  The Android communication information database file is typically stored in:
  burners.db
  """

  NAME = 'android_communication_information'
  DATA_FORMAT = 'Android communication information SQLite database file'

  REQUIRED_STRUCTURE = {
      'burners': frozenset([
          '_id', 'phone_number_id', 'voicemail_url', 'user_id', 'name', 'alias',
          'features', 'total_minutes', 'expiration_date', 'date_created',
          'last_updated_date', 'renewal_date'])
  }

  QUERIES = [(
      ('SELECT _id AS id, phone_number_id, voicemail_url, user_id, name, '
       'alias, features, total_minutes, expiration_date, date_created, '
       'last_updated_date, renewal_date FROM burners'), '_ParseBurnersRow')]

  SCHEMAS = [{
      'android_metadata': (
          'CREATE TABLE android_metadata (locale TEXT)'),
      'burners': (
          'CREATE TABLE burners(_id integer primary key autoincrement, '
          'burner_id text not null unique, phone_number_id text not null, '
          'voicemail_url text, user_id text not null, name text, alias text, '
          'features text not null, total_minutes integer not null, '
          'remaining_minutes integer not null, total_texts integer not null, '
          'remaining_texts integer not null, expiration_date integer not '
          'null, ringer integer default 1, notifications integer default 1, '
          'disabled integer default 0, date_created integer not null, '
          'last_updated_date integer, extension_count integer default 0, '
          'renewal_date integer default 0, auto_reply_active integer default '
          '0, auto_reply_text text, caller_id_enabled integer default 0, '
          'useSip integer default 0, hexColor integer, call_forward_status '
          'text)'),
      'connections': (
          'CREATE TABLE connections(_id integer primary key autoincrement, '
          'user_id integer not null, burner_id integer not null, name text, '
          'handle text, image_url text, service_name text not null, status '
          'text)'),
      'contacts': (
          'CREATE TABLE contacts(_id integer primary key autoincrement, '
          'contact_id text not null unique, name text not null, phone_number '
          'text not null, burner_id text, date_created integer not null, '
          'last_updated_date integer not null, blocked integer default 0, '
          'muted integer default 0, notes text, images text)'),
      'messages': (
          'CREATE TABLE messages(_id integer primary key autoincrement, '
          'message_id text not null unique, state integer default 0, message '
          'text not null, contact_phone_number text not null, connected '
          'integer default 0, message_type integer not null, '
          'contact_burner_id text, voice_url text, date_created integer not '
          'null, last_updated_date integer not null, read integer default 0, '
          'user_id text not null, sid text, burner_id text not null, duration '
          'integer, direction integer not null, asset_url text, send_status '
          'integer default 0 )'),
      'subscriptions': (
          'CREATE TABLE subscriptions(_id integer primary key autoincrement, '
          'subscription_id text not null unique, sku text not null, receipt '
          'text not null, burner_ids text, renewal_date integer, '
          'burner_assigned_in_period integer, store text not null,canceled '
          'integer not null, cancellation_date integer)')}]

  def _GetDateTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.JavaTime: date and time value or None if not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    if timestamp is None:
      return None

    return dfdatetime_java_time.JavaTime(timestamp=timestamp)

  def _ParseBurnersRow(self, parser_mediator, query, row, **unused_kwargs):
    """Parses a burners record row.

      Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    # TODO: parse features.
    # TODO: parse renewal_date.

    event_data = AndroidBurnerEventData()
    event_data.alias = self._GetRowValue(query_hash, row, 'alias')
    event_data.creation_time = self._GetDateTimeRowValue(
        query_hash, row, 'date_created')
    event_data.expiration_time = self._GetDateTimeRowValue(
        query_hash, row, 'expiration_date')
    event_data.last_updated_time = self._GetDateTimeRowValue(
        query_hash, row, 'last_updated_date')
    event_data.name = self._GetRowValue(query_hash, row, 'name')
    event_data.phone_number = self._GetRowValue(
        query_hash, row, 'phone_number_id')
    event_data.total_minutes = self._GetRowValue(
        query_hash, row, 'total_minutes')
    event_data.user_identifier = self._GetRowValue(query_hash, row, 'user_id')
    event_data.voicemail_url = self._GetRowValue(
        query_hash, row, 'voicemail_url')

    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(AndroidBurnerPlugin)



"""Output module for the Excel Spreadsheet (XLSX) output format."""

import datetime
import os
import re

import pytz
import xlsxwriter

from plaso.output import dynamic
from plaso.output import interface
from plaso.output import manager


class XLSXOutputModule(interface.OutputModule):
  """Output module for the Excel Spreadsheet (XLSX) output format."""

  NAME = 'xlsx'
  DESCRIPTION = 'Excel Spreadsheet (XLSX) output'

  SUPPORTS_ADDITIONAL_FIELDS = True
  SUPPORTS_CUSTOM_FIELDS = True

  WRITES_OUTPUT_FILE = True

  _DEFAULT_FIELDS = [
      'datetime', 'timestamp_desc', 'source', 'source_long',
      'message', 'parser', 'display_name', 'tag']

  _DEFAULT_TIMESTAMP_FORMAT = 'YYYY-MM-DD HH:MM:SS.000'

  _MAXIMUM_COLUMN_WIDTH = 50
  _MINIMUM_COLUMN_WIDTH = 6

  # Illegal XML string characters.
  _ILLEGAL_XML_RE = re.compile((
      r'[\x00-\x08\x0b-\x1f\x7f-\x84\x86-\x9f\ud800-\udfff\ufdd0-\ufddf'
      r'\ufffe-\uffff]'))

  def __init__(self):
    """Initializes an output module."""
    super().__init__()
    self._column_widths = []
    self._current_row = 0
    self._custom_fields = {}
    self._field_formatting_helper = dynamic.DynamicFieldFormattingHelper()
    self._field_names = self._DEFAULT_FIELDS
    self._sheet = None
    self._timestamp_format = self._DEFAULT_TIMESTAMP_FORMAT
    self._workbook = None

  def _FormatDateTime(self, output_mediator, event, event_data):  # pylint: disable=missing-return-type-doc
    """Formats the date to a datetime object without timezone information.

    Note: timezone information must be removed due to lack of support
    by xlsxwriter and Excel.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.

    Returns:
      datetime.datetime|str: date and time value or a string containing
          "ERROR" on OverflowError.
    """
    try:
      datetime_object = datetime.datetime(
          1970, 1, 1, 0, 0, 0, 0, tzinfo=pytz.UTC)
      datetime_object += datetime.timedelta(microseconds=event.timestamp)
      datetime_object.astimezone(output_mediator.time_zone)

      return datetime_object.replace(tzinfo=None)

    except (OSError, OverflowError, TypeError, ValueError) as exception:
      self._ReportEventError(event, event_data, (
          f'unable to copy timestamp: {event.timestamp!s} to a human readable '
          f'date and time with error: {exception!s}. Defaulting to: "ERROR"'))
      return 'ERROR'

  def _SanitizeField(self, field):
    """Sanitizes a field for output.

    This method replaces any illegal XML string characters with the Unicode
    replacement character (\ufffd).

    Args:
      field (str): value of the field to sanitize.

    Returns:
      str: sanitized value of the field.
    """
    return self._ILLEGAL_XML_RE.sub('\ufffd', field)

  def Close(self):
    """Closes the workbook."""
    for column_index, column_width in enumerate(self._column_widths):
      self._sheet.set_column(column_index, column_index, column_width)

    self._workbook.close()
    self._workbook = None

  def GetFieldValues(
      self, output_mediator, event, event_data, event_data_stream, event_tag):
    """Retrieves the output field values.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.
      event_tag (EventTag): event tag.

    Returns:
      dict[str, str]: output field values per name.
    """
    field_values = {}
    for field_name in self._field_names:
      if field_name == 'datetime':
        field_values['datetime'] = self._FormatDateTime(
            output_mediator, event, event_data)
        continue

      field_value = self._field_formatting_helper.GetFormattedField(
          output_mediator, field_name, event, event_data, event_data_stream,
          event_tag)

      if field_value is None and field_name in self._custom_fields:
        field_value = self._custom_fields.get(field_name)

      if field_value is None:
        field_value = '-'
      else:
        field_value = self._SanitizeField(field_value)

      field_values[field_name] = field_value

    return field_values

  def Open(self, path=None, **kwargs):  # pylint: disable=arguments-differ
    """Creates a new workbook.

    Args:
      path (Optional[str]): path of the output file.

    Raises:
      OSError: if the specified output file already exists.
      ValueError: if path is not set.
    """
    if not path:
      raise ValueError('Missing filename.')

    if os.path.isfile(path):
      raise OSError(
          f'Unable to use an already existing file for output [{path:s}]')

    options = {
        'constant_memory': True,
        'strings_to_urls': False,
        'strings_to_formulas': False,
        'default_date_format': self._timestamp_format}
    self._workbook = xlsxwriter.Workbook(path, options)
    self._sheet = self._workbook.add_worksheet('Sheet')
    self._current_row = 0

  def SetAdditionalFields(self, field_names):
    """Sets the names of additional fields to output.

    Args:
      field_names (list[str]): names of additional fields to output.
    """
    self._field_names.extend(field_names)

  def SetCustomFields(self, field_names_and_values):
    """Sets the names and values of custom fields to output.

    Args:
      field_names_and_values (list[tuple[str, str]]): names and values of
          custom fields to output.
    """
    self._custom_fields = dict(field_names_and_values)
    self._field_names.extend(self._custom_fields.keys())

  def SetFields(self, field_names):
    """Sets the names of the fields to output.

    Args:
      field_names (list[str]): names of the fields to output.
    """
    self._field_names = field_names

  def SetTimestampFormat(self, timestamp_format):
    """Set the timestamp format to use for the datetime column.

    Args:
      timestamp_format (str): format string of date and time values.
    """
    self._timestamp_format = timestamp_format

  def WriteFieldValues(self, output_mediator, field_values):
    """Writes field values to the output.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      field_values (dict[str, str]): output field values per name.
    """
    for column_index, field_name in enumerate(self._field_names):
      field_value = field_values.get(field_name)
      if field_name == 'datetime' and field_value:
        self._sheet.write_datetime(self._current_row, column_index, field_value)
        column_width = len(self._timestamp_format) + 2
      else:
        field_value = field_value or ''

        self._sheet.write(self._current_row, column_index, field_value)
        column_width = len(field_value) + 2

      # Auto adjust the column width based on the length of the output value.
      column_width = min(column_width, self._MAXIMUM_COLUMN_WIDTH)
      column_width = max(column_width, self._MINIMUM_COLUMN_WIDTH,
                         self._column_widths[column_index])

      self._column_widths[column_index] = column_width

    self._current_row += 1

  def WriteHeader(self, output_mediator):
    """Writes the header to the spreadsheet.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
    """
    cell_format = self._workbook.add_format({'bold': True})
    cell_format.set_align('center')

    self._column_widths = []
    for column_index, field_name in enumerate(self._field_names):
      self._sheet.write(0, column_index, field_name, cell_format)

      column_width = len(field_name) + 2
      self._column_widths.append(column_width)

    self._current_row = 1
    self._sheet.autofilter(0, len(self._field_names) - 1, 0, 0)
    self._sheet.freeze_panes(1, 0)


manager.OutputManager.RegisterOutput(XLSXOutputModule)



"""This file contains basic interface for plugins within Plaso.

This library serves a basis for all plugins in Plaso, whether that are Windows
Registry plugins, SQLite plugins or any other parsing plugins.

This is provided as a separate file to make it easier to inherit in other
projects that may want to use the Plaso plugin system.
"""


class BasePlugin:
  """A plugin is a lightweight parser that makes use of a common data structure.

  When a data structure is common among several artifacts or files a plugin
  infrastructure can be written to make writing parsers simpler. The goal of a
  parser plugin is to have only a single parser that understands the data
  structure that can call plugins that have specialized knowledge of certain
  structures.

  An example of this is a SQLite database. A plugin can be written that has
  knowledge of certain database, such as Chrome history, or Skype history, etc.
  This can be done without needing to write a fully-fledged parser that needs to
  re-implement the data structure knowledge. A single parser can be created that
  calls the plugins to see if it knows that particular database.

  Another example is Windows Registry, there a single parser that can parse the
  Registry can be made and the job of a single plugin is to parse a particular
  Registry key. The parser can then read a Registry key and compare it to a list
  of available plugins to see if it can be parsed.
  """

  # The name of the plugin. This is the name that is used in the registration
  # and used for parser/plugin selection, so this needs to be concise and unique
  # for all plugins/parsers, such as 'Chrome', 'Safari' or 'UserAssist'.
  NAME = 'base_plugin'

  # Data format supported by the parser plugin. This information is used by
  # the parser manager to generate parser and plugin information.
  DATA_FORMAT = ''

  # pylint: disable=unused-argument
  def Process(self, parser_mediator, **kwargs):
    """Extracts events using a parser plugin.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      kwargs (dict[str, object]): Depending on the plugin they may require
          different sets of arguments to be able to evaluate whether or not
          this is the correct plugin.

    Raises:
      ValueError: when there are unused keyword arguments.
    """
    if kwargs:
      names = ', '.join(kwargs.keys())
      raise ValueError(f'Unused keyword arguments: {names:s}')

  def UpdateChainAndProcess(self, parser_mediator, **kwargs):
    """Extracts events using a parser plugin and synchronizes the parser chain.

    This method updates the parser chain object held by the mediator, transfers
    control to the plugin-specific Process() method, and updates the chain again
    once the processing is complete.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
    """
    parser_mediator.AppendToParserChain(self.NAME)

    parser_chain = parser_mediator.GetParserChain()
    parser_mediator.SampleStartTiming(parser_chain)

    try:
      self.Process(parser_mediator, **kwargs)

    finally:
      parser_mediator.SampleStopTiming(parser_chain)

      parser_mediator.PopFromParserChain()


class BasePluginCache:
  """A generic cache for parser plugins."""

  def GetResults(self, attribute_name, default_value=None):
    """Retrieves a cached attribute.

    Args:
      attribute_name (str): name of the cached attribute.
      default_value (Optional[object]): default value.

    Returns:
      object: value of the cached attribute or default value if the cache
          does not contain the attribute.
    """
    return getattr(self, attribute_name, default_value)

  def SetResults(self, attribute_name, value):
    """Sets a cached attribute.

    Args:
      attribute_name (str): name of the cached attribute.
      value (object): value of the attribute.
    """
    setattr(self, attribute_name, value)



"""Fake (in-memory only) store for testing."""

from acstore import fake_store as acstore_fake_store

from plaso.containers import events
from plaso.storage.fake import event_heap


class FakeStore(acstore_fake_store.FakeAttributeContainerStore):
  """Fake (in-memory only) store for testing.

  Attributes:
    serialization_format (str): serialization format.
  """

  _CONTAINER_TYPE_EVENT = events.EventObject.CONTAINER_TYPE

  def __init__(self):
    """Initializes a fake (in-memory only) store."""
    super().__init__()
    self._serializers_profiler = None
    self.serialization_format = None

  def GetSortedEvents(self, time_range=None):
    """Retrieves the events in increasing chronological order.

    Args:
      time_range (Optional[TimeRange]): time range used to filter events
          that fall in a specific period.

    Returns:
      generator(EventObject): event generator.

    Raises:
      OSError: when the storage writer is closed.
    """
    if not self._is_open:
      raise OSError('Unable to read from closed storage writer.')

    generator = self.GetAttributeContainers(self._CONTAINER_TYPE_EVENT)
    sorted_events = event_heap.EventHeap()

    for event_index, event in enumerate(generator):
      if (time_range and (
          event.timestamp < time_range.start_timestamp or
          event.timestamp > time_range.end_timestamp)):
        continue

      # The event index is used to ensure to sort events with the same date and
      # time and description in the order they were added to the store.
      sorted_events.PushEvent(event, event_index)

    return iter(sorted_events.PopEvents())

  def SetSerializersProfiler(self, serializers_profiler):
    """Sets the serializers profiler.

    Args:
      serializers_profiler (SerializersProfiler): serializers profiler.
    """
    self._serializers_profiler = serializers_profiler



"""Windows Shortcut (LNK) custom event formatter helpers."""

from plaso.formatters import interface
from plaso.formatters import manager


class WindowsShortcutLinkedPathFormatterHelper(
    interface.CustomEventFormatterHelper):
  """Windows Shortcut (LNK) linked path formatter helper."""

  IDENTIFIER = 'windows_shortcut_linked_path'

  def FormatEventValues(self, output_mediator, event_values):
    """Formats event values using the helper.

    Args:
      output_mediator (OutputMediator): output mediator.
      event_values (dict[str, object]): event values.
    """
    linked_path = event_values.get('local_path')
    if not linked_path:
      linked_path = event_values.get('network_path')

    if not linked_path:
      linked_path = event_values.get('relative_path')
      if linked_path:
        working_directory = event_values.get('working_directory')
        if working_directory:
          linked_path = '\\'.join([working_directory, linked_path])

    event_values['linked_path'] = linked_path or 'Unknown'


manager.FormattersManager.RegisterEventFormatterHelper(
    WindowsShortcutLinkedPathFormatterHelper)



"""SQLite parser plugin for iOS accounts (Accounts3.db) database files."""

from dfdatetime import cocoa_time as dfdatetime_cocoa_time

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class IOSAccountsEventData(events.EventData):
  """iOS accounts event data.

  Attributes:
    account_type (str): account type.
    creation_time (dfdatetime.DateTimeValues): date and time the account
        was created.
    identifier (str): identifier.
    owning_bundle_identifier (str): owning bundle identifier of the
        application managing the account.
    username (str): user name.
  """

  DATA_TYPE = 'ios:accounts:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.account_type = None
    self.creation_time = None
    self.identifier = None
    self.owning_bundle_identifier = None
    self.username = None


class IOSAccountsPlugin(interface.SQLitePlugin):
  """SQLite parser plugin for iOS accounts (Accounts3.db) database files."""

  NAME = 'ios_accounts'
  DATA_FORMAT = 'iOS accounts SQLite database (Accounts3.db) file'

  REQUIRED_STRUCTURE = {
      'ZACCOUNT': frozenset([
          'ZACCOUNTTYPE', 'ZDATE', 'ZUSERNAME', 'ZIDENTIFIER',
          'ZOWNINGBUNDLEID']),
      'ZACCOUNTTYPE': frozenset([
          'Z_PK', 'ZACCOUNTTYPEDESCRIPTION'])}

  QUERIES = [(
      ('SELECT ZACCOUNT.ZDATE, ZACCOUNTTYPE.ZACCOUNTTYPEDESCRIPTION, '
       'ZACCOUNT.ZUSERNAME, ZACCOUNT.ZIDENTIFIER, ZACCOUNT.ZOWNINGBUNDLEID '
       'FROM ZACCOUNT LEFT JOIN ZACCOUNTTYPE '
       'ON ZACCOUNT.ZACCOUNTTYPE = ZACCOUNTTYPE.Z_PK'),
      'ParseAccountRow')]

  SCHEMAS = [{
      'ZACCESSOPTIONSKEY': (
          'CREATE TABLE ZACCESSOPTIONSKEY ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZNAME VARCHAR )'),
      'ZACCOUNT': (
          'CREATE TABLE ZACCOUNT ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
          'Z_OPT INTEGER, ZACTIVE INTEGER, ZAUTHENTICATED INTEGER, '
          'ZSUPPORTSAUTHENTICATION INTEGER, ZVISIBLE INTEGER, ZACCOUNTTYPE '
          'INTEGER, ZPARENTACCOUNT INTEGER, ZDATE TIMESTAMP, '
          'ZLASTCREDENTIALRENEWALREJECTIONDATE TIMESTAMP, ZACCOUNTDESCRIPTION '
          'VARCHAR, ZAUTHENTICATIONTYPE VARCHAR, ZCREDENTIALTYPE VARCHAR, '
          'ZIDENTIFIER VARCHAR, ZOWNINGBUNDLEID VARCHAR, ZUSERNAME VARCHAR, '
          'ZDATACLASSPROPERTIES BLOB )'),
      'ZACCOUNTPROPERTY': (
          'CREATE TABLE ZACCOUNTPROPERTY ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZOWNER INTEGER, ZKEY VARCHAR, ZVALUE BLOB '
          ')'),
      'ZACCOUNTTYPE': (
          'CREATE TABLE ZACCOUNTTYPE ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZENCRYPTACCOUNTPROPERTIES INTEGER, '
          'ZOBSOLETE INTEGER, ZSUPPORTSAUTHENTICATION INTEGER, '
          'ZSUPPORTSMULTIPLEACCOUNTS INTEGER, ZVISIBILITY INTEGER, '
          'ZACCOUNTTYPEDESCRIPTION VARCHAR, ZCREDENTIALPROTECTIONPOLICY '
          'VARCHAR, ZCREDENTIALTYPE VARCHAR, ZIDENTIFIER VARCHAR, '
          'ZOWNINGBUNDLEID VARCHAR )'),
      'ZAUTHORIZATION': (
          'CREATE TABLE ZAUTHORIZATION ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZACCOUNTTYPE INTEGER, ZBUNDLEID VARCHAR, '
          'ZGRANTEDPERMISSIONS VARCHAR, ZOPTIONS BLOB )'),
      'ZCREDENTIALITEM': (
          'CREATE TABLE ZCREDENTIALITEM ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZPERSISTENT INTEGER, ZEXPIRATIONDATE '
          'TIMESTAMP, ZACCOUNTIDENTIFIER VARCHAR, ZSERVICENAME VARCHAR )'),
      'ZDATACLASS': (
          'CREATE TABLE ZDATACLASS ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
          'Z_OPT INTEGER, ZNAME BLOB )'),
      'Z_1OWNINGACCOUNTTYPES': (
          'CREATE TABLE Z_1OWNINGACCOUNTTYPES ( Z_1ACCESSKEYS INTEGER, '
          'Z_4OWNINGACCOUNTTYPES INTEGER, PRIMARY KEY (Z_1ACCESSKEYS, '
          'Z_4OWNINGACCOUNTTYPES) )'),
      'Z_2ENABLEDDATACLASSES': (
          'CREATE TABLE Z_2ENABLEDDATACLASSES ( Z_2ENABLEDACCOUNTS INTEGER, '
          'Z_7ENABLEDDATACLASSES INTEGER, PRIMARY KEY (Z_2ENABLEDACCOUNTS, '
          'Z_7ENABLEDDATACLASSES) )'),
      'Z_2PROVISIONEDDATACLASSES': (
          'CREATE TABLE Z_2PROVISIONEDDATACLASSES ( Z_2PROVISIONEDACCOUNTS '
          'INTEGER, Z_7PROVISIONEDDATACLASSES INTEGER, PRIMARY KEY '
          '(Z_2PROVISIONEDACCOUNTS, Z_7PROVISIONEDDATACLASSES) )'),
      'Z_4SUPPORTEDDATACLASSES': (
          'CREATE TABLE Z_4SUPPORTEDDATACLASSES ( Z_4SUPPORTEDTYPES INTEGER, '
          'Z_7SUPPORTEDDATACLASSES INTEGER, PRIMARY KEY (Z_4SUPPORTEDTYPES, '
          'Z_7SUPPORTEDDATACLASSES) )'),
      'Z_4SYNCABLEDATACLASSES': (
          'CREATE TABLE Z_4SYNCABLEDATACLASSES ( Z_4SYNCABLETYPES INTEGER, '
          'Z_7SYNCABLEDATACLASSES INTEGER, PRIMARY KEY (Z_4SYNCABLETYPES, '
          'Z_7SYNCABLEDATACLASSES) )'),
      'Z_METADATA': (
          'CREATE TABLE Z_METADATA (Z_VERSION INTEGER PRIMARY KEY, Z_UUID '
          'VARCHAR(255), Z_PLIST BLOB)'),
      'Z_MODELCACHE': (
          'CREATE TABLE Z_MODELCACHE (Z_CONTENT BLOB)'),
      'Z_PRIMARYKEY': (
          'CREATE TABLE Z_PRIMARYKEY (Z_ENT INTEGER PRIMARY KEY, Z_NAME '
          'VARCHAR, Z_SUPER INTEGER, Z_MAX INTEGER)')}]

  REQUIRES_SCHEMA_MATCH = False

  def _GetTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely
          identifies the query that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.CocoaTime: date and time value or None if not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    if timestamp is None:
      return None

    return dfdatetime_cocoa_time.CocoaTime(timestamp=timestamp)

  # pylint: disable=unused-argument
  def ParseAccountRow(self, parser_mediator, query, row, **unused_kwargs):
    """Parses an account row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between
          parsers and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    event_data = IOSAccountsEventData()
    event_data.account_type = self._GetRowValue(
        query_hash, row, 'ZACCOUNTTYPEDESCRIPTION')
    event_data.creation_time = self._GetTimeRowValue(query_hash, row, 'ZDATE')
    event_data.identifier = self._GetRowValue(
        query_hash, row, 'ZIDENTIFIER')
    event_data.owning_bundle_identifier = self._GetRowValue(
        query_hash, row, 'ZOWNINGBUNDLEID')
    event_data.username = self._GetRowValue(query_hash, row, 'ZUSERNAME')

    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(IOSAccountsPlugin)



"""This file contains a parser for compound ZIP files."""

import zipfile

from plaso.lib import errors
from plaso.parsers import interface
from plaso.parsers import logger
from plaso.parsers import manager


class CompoundZIPParser(interface.FileObjectParser):
  """Shared functionality for parsing compound ZIP files.

  Compound ZIP files are ZIP files used as containers to create another file
  format, as opposed to archives of unrelated files.
  """

  NAME = 'czip'
  DATA_FORMAT = 'Compound ZIP file'

  _plugin_classes = {}

  def ParseFileObject(self, parser_mediator, file_object):
    """Parses a compound ZIP file-like object.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      file_object (dfvfs.FileIO): a file-like object.

    Raises:
      WrongParser: when the file cannot be parsed.
    """
    display_name = parser_mediator.GetDisplayName()

    if not zipfile.is_zipfile(file_object):
      raise errors.WrongParser(
          f'[{self.NAME:s}] unable to parse file: {display_name:s} with error: '
          f'Not a Zip file.')

    try:
      zip_file = zipfile.ZipFile(file_object, 'r', allowZip64=True)  # pylint: disable=consider-using-with

    # Some non-ZIP files return true for is_zipfile but will fail with another
    # error like a negative seek (OSError). Note that this function can raise
    # many different exceptions.
    except Exception as exception:  # pylint: disable=broad-except
      raise errors.WrongParser(
          f'[{self.NAME:s}] unable to parse file: {display_name:s} '
          f'with error: {exception!s}')

    for plugin_name, plugin in self._plugins_per_name.items():
      if parser_mediator.abort:
        break

      profiling_name = '/'.join([self.NAME, plugin.NAME])

      parser_mediator.SampleFormatCheckStartTiming(profiling_name)

      try:
        result = plugin.CheckRequiredPaths(zip_file)
      finally:
        parser_mediator.SampleFormatCheckStopTiming(profiling_name)

      if not result:
        logger.debug(
            f'Skipped parsing file: {display_name:s} with plugin: '
            f'{plugin_name:s}')
        continue

      logger.debug(
          f'Parsing file: {display_name:s} with plugin: {plugin_name:s}')

      parser_mediator.SampleStartTiming(profiling_name)

      try:
        plugin.UpdateChainAndProcess(parser_mediator, zip_file=zip_file)

      except Exception as exception:  # pylint: disable=broad-except
        parser_mediator.ProduceExtractionWarning(
            f'plugin: {plugin_name:s} unable to parse ZIP file: '
            f'{display_name:s} with error: {exception!s}')

      finally:
        parser_mediator.SampleStopTiming(profiling_name)

    zip_file.close()


manager.ParsersManager.RegisterParser(CompoundZIPParser)



"""Windows known folders helper."""


class WindowsKnownFoldersHelper:
  """Windows known folders helper."""

  # pylint: disable=line-too-long

  _PATH_PER_GUID = {
      '{008ca0b1-55b4-4c56-b8a8-4de4b299d3be}': '{%APPDATA%\\Microsoft\\Windows\\AccountPictures',
      '{00bcfc5a-ed94-4e48-96a1-3f6217f21990}': '{%LOCALAPPDATA%\\Microsoft\\Windows\\RoamingTiles',
      '{0139d44e-6afe-49f2-8690-3dafcae6ffb8}': '{%ALLUSERSPROFILE%\\Microsoft\\Windows\\Start Menu\\Programs',
      '{0482af6c-08f1-4c34-8c90-e17ec98b1e17}': '%PUBLIC%\\AccountPictures',
      '{054fae61-4dd8-4787-80b6-090220c4b700}': '{%LOCALAPPDATA%\\Microsoft\\Windows\\GameExplorer',
      '{0762d272-c50a-4bb0-a382-697dcd729b80}': '%SYSTEMDRIVE%\\Users',
      '{0d4c3db6-03a3-462f-a0e6-08924c41b5d4}': '{%LOCALAPPDATA%\\Microsoft\\Windows\\ConnectedSearch\\History',
      '{15ca69b3-30ee-49c1-ace1-6b5ec372afb5}': '{%PUBLIC%\\Music\\Sample Playlists',
      '{1777f761-68ad-4d8a-87bd-30b759fa33dd}': '%USERPROFILE%\\Favorites',
      '{18989b1d-99b5-455b-841c-ab7c74e4ddfc}': '%USERPROFILE%\\Videos',
      '{1a6fdba2-f42d-4358-a798-b74d745926c5}': '%PUBLIC%\\RecordedTV.library-ms',
      '{1ac14e77-02e7-4e5d-b744-2eb1ae5198b7}': '%WINDIR%\\System32',
      '{1b3ea5dc-b587-4786-b4ef-bd1dc332aeae}': '{%APPDATA%\\Microsoft\\Windows\\Libraries',
      '{2112ab0a-c86a-4ffe-a368-0de96e47012e}': '{%APPDATA%\\Microsoft\\Windows\\Libraries\\Music.library-ms',
      '{2400183a-6185-49fb-a2d8-4a392a602ba3}': '%PUBLIC%\\Videos',
      '{24d89e24-2f19-4534-9dde-6a6671fbb8fe}': '{%USERPROFILE%\\OneDrive\\Documents',
      '{2a00375e-224c-49de-b8d1-440df7ef3ddc}': '%WINDIR%\\resources\\%CODEPAGE%',
      '{2b0f765d-c0e9-4171-908e-08a611b84ff6}': '{%APPDATA%\\Microsoft\\Windows\\Cookies',
      '{2c36c0aa-5812-4b87-bfd0-4cd0dfb19b39}': '{%LOCALAPPDATA%\\Microsoft\\Windows Photo Gallery\\Original Images',
      '{3214fab5-9757-4298-bb61-92a9deaa44ff}': '%PUBLIC%\\Music',
      '{339719b5-8c47-4894-94c2-d8f77add44a6}': '{%USERPROFILE%\\OneDrive\\Pictures',
      '{33e28130-4e1e-4676-835a-98395c3bc3bb}': '%USERPROFILE%\\Pictures',
      '{352481e8-33be-4251-ba85-6007caedcf9d}': '{%LOCALAPPDATA%\\Microsoft\\Windows\\Temporary Internet Files',
      '{374de290-123f-4565-9164-39c4925e467b}': '%USERPROFILE%\\Downloads',
      '{3d644c9b-1fb8-4f30-9b45-f670235f79c0}': '%PUBLIC%\\Downloads',
      '{3eb685db-65f9-4cf6-a03a-e3ef65729f3d}': '%USERPROFILE%\\AppData\\Roaming',
      '{48daf80b-e6cf-4f4e-b800-0e69d84ee384}': '{%ALLUSERSPROFILE%\\Microsoft\\Windows\\Libraries',
      '{491e922f-5643-4af4-a7eb-4e7a138d8174}': '{%APPDATA%\\Microsoft\\Windows\\Libraries\\Videos.library-ms',
      '{4bd8d571-6d19-48d3-be97-422220080e43}': '%USERPROFILE%\\Music',
      '{4c5c32ff-bb9d-43b0-b5b4-2d72e54eaaa4}': '%USERPROFILE%\\Saved Games',
      '{52a4f021-7b75-48a9-9f6b-4b87a210bc8f}': '{%APPDATA%\\Microsoft\\Internet Explorer\\Quick Launch',
      '{5cd7aee2-2219-4a67-b85d-6c9ce15660cb}': '%LOCALAPPDATA%\\Programs',
      '{5ce4a5e9-e4eb-479d-b89f-130c02886155}': '{%ALLUSERSPROFILE%\\Microsoft\\Windows\\DeviceMetadataStore',
      '{5e6c858f-0e22-4760-9afe-ea3317b67173}': '{%SYSTEMDRIVE%\\Users\\%USERNAME%',
      '{625b53c3-ab48-4ec1-ba1f-a1ef4146fc19}': '{%APPDATA%\\Microsoft\\Windows\\Start Menu',
      '{62ab5d82-fdc1-4dc3-a9dd-070d1d495d97}': '%SYSTEMDRIVE%\\ProgramData',
      '{6365d5a7-0f0d-45e5-87f6-0da56b6a4f7d}': '%PROGRAMFILES%\\Common Files',
      '{69d2cf90-fc33-4fb7-9a0c-ebb0f0fcb43c}': '{%USERPROFILE%\\Pictures\\Slide Shows',
      '{6d809377-6af0-444b-8957-a3773f02200e}': '%SYSTEMDRIVE%\\Program Files',
      '{724ef170-a42d-4fef-9f26-b60e846fba4f}': '{%APPDATA%\\Microsoft\\Windows\\Start Menu\\Programs\\Administrative Tools',
      '{767e6811-49cb-4273-87c2-20f355e1085b}': '%USERPROFILE%\\OneDrive\\Pictures\\Camera Roll',
      '{7b0db17d-9cd2-4a93-9733-46cc89022e7c}': '%APPDATA%\\Microsoft\\Windows\\Libraries\\Documents.library-ms',
      '{7b396e54-9ec5-4300-be0a-2482ebae1a26}': '%PROGRAMFILES%\\Windows Sidebar\\Gadgets',
      '{7c5a40ef-a0fb-4bfc-874a-c0f2e0b9fa8e}': '%PROGRAMFILES% (%SYSTEMDRIVE%\\Program Files)',
      '{7d1d3a04-debb-4115-95cf-2f29da2920da}': '%USERPROFILE%\\Searches',
      '{7e636bfe-dfa9-4d5e-b456-d7b39851d8a9}': '%LOCALAPPDATA%\\Microsoft\\Windows\\ConnectedSearch\\Templates',
      '{82a5ea35-d9cd-47c5-9629-e15d2f714e6e}': '%ALLUSERSPROFILE%\\Microsoft\\Windows\\Start Menu\\Programs\\StartUp',
      '{859ead94-2e85-48ad-a71a-0969cb56a6cd}': '%PUBLIC%\\Videos\\Sample Videos',
      '{8983036c-27c0-404b-8f08-102d10dcfd74}': '%APPDATA%\\Microsoft\\Windows\\SendTo',
      '{8ad10c31-2adb-4296-a8f7-e4701232c972}': '%WINDIR%\\Resources',
      '{905e63b6-c1bf-494e-b29c-65b732d3d21a}': '%SYSTEMDRIVE%\\Program Files',
      '{9274bd8d-cfd1-41c3-b35e-b13f55a758f4}': '%APPDATA%\\Microsoft\\Windows\\Printer Shortcuts',
      '{9e3995ab-1f9c-4f13-b827-48b24b6c7174}': '%APPDATA%\\Microsoft\\Internet Explorer\\Quick Launch\\User Pinned',
      '{9e52ab10-f80d-49df-acb8-4330f5687855}': '%LOCALAPPDATA%\\Microsoft\\Windows\\Burn\\Burn',
      '{a3918781-e5f2-4890-b3d9-a7e54332328c}': '%LOCALAPPDATA%\\Microsoft\\Windows\\Application Shortcuts',
      '{a4115719-d62e-491d-aa7c-e74b8be3b067}': '%ALLUSERSPROFILE%\\Microsoft\\Windows\\Start Menu',
      '{a520a1a4-1780-4ff6-bd18-167343c5af16}': '%USERPROFILE%\\AppData\\LocalLow',
      '{a52bba46-e9e1-435f-b3d9-28daa648c0f6}': '%USERPROFILE%\\OneDrive',
      '{a63293e8-664e-48db-a079-df759e0509f7}': '%APPDATA%\\Microsoft\\Windows\\Templates',
      '{a75d362e-50fc-4fb7-ac2c-a8beaa314493}': '%LOCALAPPDATA%\\Microsoft\\Windows Sidebar\\Gadgets',
      '{a77f5d77-2e2b-44c3-a6a2-aba601054a51}': '%APPDATA%\\Microsoft\\Windows\\Start Menu\\Programs',
      '{a990ae9f-a03b-4e80-94bc-9912d7504104}': '%APPDATA%\\Microsoft\\Windows\\Libraries\\Pictures.library-ms',
      '{aaa8d5a5-f1d6-4259-baa8-78e7ef60835e}': '%LOCALAPPDATA%\\Microsoft\\Windows\\RoamedTileImages',
      '{ab5fb87b-7ce2-4f83-915d-550846c9537b}': '%USERPROFILE%\\Pictures\\Camera Roll',
      '{ae50c081-ebd2-438a-8655-8a092e34987a}': '%APPDATA%\\Microsoft\\Windows\\Recent',
      '{b250c668-f57d-4ee1-a63c-290ee7d1aa1f}': '%PUBLIC%\\Music\\Sample Music',
      '{b4bfcc3a-db2c-424c-b029-7fe99a87c641}': '%USERPROFILE%\\Desktop',
      '{b6ebfb86-6907-413c-9af7-4fc2abf07cc5}': '%PUBLIC%\\Pictures',
      '{b7bede81-df94-4682-a7d8-57a52620b86f}': '%USERPROFILE%\\Pictures\\Screenshots',
      '{b94237e7-57ac-4347-9151-b08c6c32d1f7}': '%ALLUSERSPROFILE%\\Microsoft\\Windows\\Templates',
      '{b97d20bb-f46a-4c97-ba10-5e3608430854}': '%APPDATA%\\Microsoft\\Windows\\Start Menu\\Programs\\StartUp',
      '{bcb5256f-79f6-4cee-b725-dc34e402fd46}': '%APPDATA%\\Microsoft\\Internet Explorer\\Quick Launch\\User Pinned\\ImplicitAppShortcuts',
      '{bcbd3057-ca5c-4622-b42d-bc56db0ae516}': '%LOCALAPPDATA%\\Programs\\Common',
      '{bfb9d5e0-c6a9-404c-b2b2-ae6db6af4968}': '%USERPROFILE%\\Links',
      '{c1bae2d0-10df-4334-bedd-7aa20b227a9d}': '%ALLUSERSPROFILE%\\OEM Links',
      '{c4900540-2379-4c75-844b-64e6faf8716b}': '%PUBLIC%\\Pictures\\Sample Pictures',
      '{c4aa340d-f20f-4863-afef-f87ef2e6ba25}': '%PUBLIC%\\Desktop',
      '{c5abbf53-e17f-4121-8900-86626fc2c973}': '%APPDATA%\\Microsoft\\Windows\\Network Shortcuts',
      '{c870044b-f49e-4126-a9c3-b52a1ff411e8}': '%LOCALAPPDATA%\\Microsoft\\Windows\\Ringtones',
      '{d0384e7d-bac3-4797-8f14-cba229b392b5}': '%ALLUSERSPROFILE%\\Microsoft\\Windows\\Start Menu\\Programs\\Administrative Tools',
      '{d65231b0-b2f1-4857-a4ce-a8e7c6ea7d27}': '%WINDIR%\\system32',
      '{d9dc8a3b-b784-432e-a781-5a1130a75963}': '%LOCALAPPDATA%\\Microsoft\\Windows\\History',
      '{de92c1c7-837f-4f69-a3bb-86e631204a23}': '%USERPROFILE%\\Music\\Playlists',
      '{de974d24-d9c6-4d3e-bf91-f4455120b917}': '%PROGRAMFILES%\\Common Files',
      '{debf2536-e1a8-4c59-b6a2-414586476aea}': '%ALLUSERSPROFILE%\\Microsoft\\Windows\\GameExplorer',
      '{dfdf76a2-c82a-4d63-906a-5644ac457385}': '%SYSTEMDRIVE%\\Users\\Public',
      '{e555ab60-153b-4d17-9f04-a5fe99fc15ec}': '%ALLUSERSPROFILE%\\Microsoft\\Windows\\Ringtones',
      '{ed4824af-dce4-45a8-81e2-fc7965083634}': '%PUBLIC%\\Documents',
      '{f1b32785-6fba-4fcf-9d55-7b8e7f157091}': '%USERPROFILE%\\AppData\\Local',
      '{f38bf404-1d43-42f2-9305-67de0b28fc23}': '%WINDIR%',
      '{f7f1ed05-9f6d-47a2-aaae-29d317c6f066}': '%PROGRAMFILES%\\Common Files',
      '{fd228cb7-ae11-4ae3-864c-16f3910ab8fe}': '%WINDIR%\\Fonts',
      '{fdd39ad0-238f-46af-adb4-6c85480369c7}': '%USERPROFILE%\\Documents'}

  # pylint: enable=line-too-long

  @classmethod
  def GetPath(cls, known_folder_identifier):
    """Retrieves the path for a specific known folder identifier.

    Args:
      known_folder_identifier (str): known folder identifier in the format
          "{GUID}".

    Returns:
      str: path represented by the known folder identifier or None of not
          available.
    """
    return cls._PATH_PER_GUID.get(known_folder_identifier.lower())



"""SQLite parser plugin for Safari history database files."""

from dfdatetime import cocoa_time as dfdatetime_cocoa_time

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class SafariHistoryPageVisitedEventData(events.EventData):
  """Safari history event data.

  Attributes:
    host (str): hostname of the server.
    last_visited_time (dfdatetime.DateTimeValues): date and time the URL was
        last visited.
    offset (str): identifier of the row, from which the event data was
        extracted.
    query (str): SQL query that was used to obtain the event data.
    title (str): title of the webpage visited.
    url (str): URL visited.
    visit_count (int): number of times the website was visited.
    was_http_non_get (bool): True if the webpage was visited using a
        non-GET HTTP request.
  """

  DATA_TYPE = 'safari:history:visit_sqlite'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.host = None
    self.last_visited_time = None
    self.offset = None
    self.query = None
    self.title = None
    self.url = None
    self.visit_count = None
    self.visit_redirect_source = None
    self.was_http_non_get = None


class SafariHistoryPluginSqlite(interface.SQLitePlugin):
  """SQLite parser plugin for Safari history database files.

  The Safari history database file is typically stored in:
  History.db
  """

  NAME = 'safari_historydb'
  DATA_FORMAT = 'Safari history SQLite database (History.db) file'

  REQUIRED_STRUCTURE = {
      'history_items': frozenset([
          'id', 'url', 'visit_count']),
      'history_visits': frozenset([
          'id', 'history_item', 'visit_time', 'redirect_destination', 'title',
          'http_non_get', 'redirect_source'])}

  QUERIES = [
      (('SELECT history_items.id, history_items.url, history_items.visit'
        '_count, history_visits.id AS visit_id, history_visits.history_item,'
        'history_visits.visit_time, history_visits.redirect_destination, '
        'history_visits.title, history_visits.http_non_get, '
        'history_visits.redirect_source '
        'FROM history_items, history_visits '
        'WHERE history_items.id = history_visits.history_item'),
       '_ParsePageVisitRow')]

  SCHEMAS = [{
      'history_client_versions': (
          'CREATE TABLE history_client_versions (client_version INTEGER '
          'PRIMARY KEY,last_seen REAL NOT NULL)'),
      'history_event_listeners': (
          'CREATE TABLE history_event_listeners (listener_name TEXT PRIMARY '
          'KEY NOT NULL UNIQUE,last_seen REAL NOT NULL)'),
      'history_events': (
          'CREATE TABLE history_events (id INTEGER PRIMARY KEY '
          'AUTOINCREMENT,event_type TEXT NOT NULL,event_time REAL NOT '
          'NULL,pending_listeners TEXT NOT NULL,value BLOB)'),
      'history_items': (
          'CREATE TABLE history_items (id INTEGER PRIMARY KEY '
          'AUTOINCREMENT,url TEXT NOT NULL UNIQUE,domain_expansion TEXT '
          'NULL,visit_count INTEGER NOT NULL,daily_visit_counts BLOB NOT '
          'NULL,weekly_visit_counts BLOB NULL,autocomplete_triggers BLOB '
          'NULL,should_recompute_derived_visit_counts INTEGER NOT '
          'NULL,visit_count_score INTEGER NOT NULL)'),
      'history_tombstones': (
          'CREATE TABLE history_tombstones (id INTEGER PRIMARY KEY '
          'AUTOINCREMENT,start_time REAL NOT NULL,end_time REAL NOT NULL,url '
          'TEXT,generation INTEGER NOT NULL DEFAULT 0)'),
      'history_visits': (
          'CREATE TABLE history_visits (id INTEGER PRIMARY KEY '
          'AUTOINCREMENT,history_item INTEGER NOT NULL REFERENCES '
          'history_items(id) ON DELETE CASCADE,visit_time REAL NOT NULL,title '
          'TEXT NULL,load_successful BOOLEAN NOT NULL DEFAULT 1,http_non_get '
          'BOOLEAN NOT NULL DEFAULT 0,synthesized BOOLEAN NOT NULL DEFAULT '
          '0,redirect_source INTEGER NULL UNIQUE REFERENCES '
          'history_visits(id) ON DELETE CASCADE,redirect_destination INTEGER '
          'NULL UNIQUE REFERENCES history_visits(id) ON DELETE CASCADE,origin '
          'INTEGER NOT NULL DEFAULT 0,generation INTEGER NOT NULL DEFAULT '
          '0,attributes INTEGER NOT NULL DEFAULT 0,score INTEGER NOT NULL '
          'DEFAULT 0)'),
      'metadata': (
          'CREATE TABLE metadata (key TEXT NOT NULL UNIQUE, value)')}]

  def _GetDateTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.CocoaTime: date and time value or None if not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    if timestamp is None:
      return None

    return dfdatetime_cocoa_time.CocoaTime(timestamp=timestamp)

  def _ParsePageVisitRow(self, parser_mediator, query, row, **unused_kwargs):
    """Parses a visited row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    was_http_non_get = self._GetRowValue(query_hash, row, 'http_non_get')

    event_data = SafariHistoryPageVisitedEventData()
    event_data.last_visited_time = self._GetDateTimeRowValue(
        query_hash, row, 'visit_time')
    event_data.offset = self._GetRowValue(query_hash, row, 'id')
    event_data.query = query
    event_data.title = self._GetRowValue(query_hash, row, 'title') or None
    event_data.url = self._GetRowValue(query_hash, row, 'url')
    event_data.visit_count = self._GetRowValue(query_hash, row, 'visit_count')
    event_data.was_http_non_get = bool(was_http_non_get)

    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(SafariHistoryPluginSqlite)



"""Plist parser plugin for Apple Account plist files.

Fields within the plist key: com.apple.coreservices.appleidauthenticationinfo
  Accounts: account name.
  FirstName: first name associated with the account.
  LastName: last (or family) name associate with the account.
  CreationDate: timestamp when the account was configured in the system.
  LastSuccessfulConnect: last time when the account was connected.
  ValidationDate: last time when the account was validated.
"""

from plaso.containers import events
from plaso.parsers import plist
from plaso.parsers.plist_plugins import interface


class AppleAccountEventData(events.EventData):
  """Apple account event data.

  Attributes:
    account_name (str): name of the account.
    creation_time (dfdatetime.DateTimeValues): date and time the Apple account
        was created (configured) on the system.
    first_name (str): first name.
    last_connected_time (dfdatetime.DateTimeValues): last date and time
        the system successfully connected to the Apple account.
    last_name (str): last (or family) name.
    validation_time (dfdatetime.DateTimeValues): date and time the Apple account
        was validated.
  """

  DATA_TYPE = 'macos:apple_account:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.account_name = None
    self.creation_time = None
    self.first_name = None
    self.last_connected_time = None
    self.last_name = None
    self.validation_time = None


class AppleAccountPlistPlugin(interface.PlistPlugin):
  """Plist parser plugin for Apple Account plist files."""

  NAME = 'apple_id'
  DATA_FORMAT = 'Apple account information plist file'

  PLIST_PATH_FILTERS = frozenset([
      interface.PrefixPlistPathFilter(
          'com.apple.coreservices.appleidauthenticationinfo')])

  PLIST_KEYS = frozenset(['AuthCertificates', 'AccessorVersions', 'Accounts'])

  # pylint: disable=arguments-differ
  def _ParsePlist(self, parser_mediator, match=None, **unused_kwargs):
    """Extracts relevant Apple Account entries.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      match (Optional[dict[str: object]]): keys extracted from PLIST_KEYS.
    """
    for account_name, plist_key in match.get('Accounts', {}).items():
      event_data = AppleAccountEventData()
      event_data.account_name = account_name
      event_data.creation_time = self._GetDateTimeValueFromPlistKey(
          plist_key, 'CreationDate')
      event_data.first_name = plist_key.get('FirstName')
      event_data.last_connected_time = self._GetDateTimeValueFromPlistKey(
          plist_key, 'LastSuccessfulConnect')
      event_data.last_name = plist_key.get('LastName')
      event_data.validation_time = self._GetDateTimeValueFromPlistKey(
          plist_key, 'ValidationDate')

      parser_mediator.ProduceEventData(event_data)


plist.PlistParser.RegisterPlugin(AppleAccountPlistPlugin)



"""SQLite-based storage reader."""

from plaso.storage import reader
from plaso.storage.sqlite import sqlite_file


class SQLiteStorageReader(reader.StorageReader):
  """SQLite-based storage reader."""

  def __init__(self, path):
    """Initializes a storage reader.

    Args:
      path (str): path to the input SQLite database.
    """
    super().__init__()
    self._path = path
    self._store = sqlite_file.SQLiteStorageFile()
    self._store.Open(path=path)



"""Parser for bencoded files."""

import os
import re

import bencode

from dfdatetime import posix_time as dfdatetime_posix_time

from plaso.lib import errors
from plaso.parsers import interface
from plaso.parsers import logger
from plaso.parsers import manager


class BencodeValues:
  """Bencode values."""

  def __init__(self, decoded_values):
    """Initializes bencode values.

    Args:
      decoded_values (collections.OrderedDict[bytes|str, object]): decoded
          values.
    """
    super().__init__()
    self._decoded_values = decoded_values

  def GetDateTimeValue(self, name):
    """Retrieves a date and time value.

    Args:
      name (str): name of the value.

    Returns:
      dfdatetime.PosixTime: date and time or None if not available.
    """
    timestamp = self.GetDecodedValue(name)
    if not timestamp:
      return None

    return dfdatetime_posix_time.PosixTime(timestamp=timestamp)

  def GetDecodedValue(self, name):
    """Retrieves a decoded value.

    Args:
      name (str): name of the value.

    Returns:
      object: decoded value or None if not available.
    """
    value = self._decoded_values.get(name)
    if value is None:
      # Work-around for issue in bencode 3.0.1 where keys are bytes.
      name_as_byte_stream = name.encode('utf-8')
      value = self._decoded_values.get(name_as_byte_stream)

    if isinstance(value, bytes):
      # Work-around for issue in bencode 3.0.1 where string values are bytes.
      value = value.decode('utf-8')

    return value

  def GetValues(self):
    """Retrieves the values.

    Yields:
      tuple[str, object]: name and decoded value.
    """
    for key, value in self._decoded_values.items():
      if isinstance(key, bytes):
        # Work-around for issue in bencode 3.0.1 where keys are bytes.
        key = key.decode('utf-8')

      yield key, value


class BencodeFile:
  """Bencode file."""

  def __init__(self):
    """Initializes a bencode file."""
    super().__init__()
    self._decoded_values = None
    self._key_names = set()

  @property
  def keys(self):
    """Set[str]: names of all the keys."""
    return self._key_names

  def Close(self):
    """Closes the file."""
    self._decoded_values = None

  def GetValues(self):
    """Retrieves the values in the root of the bencode file.

    Returns:
      BencodeValues: values.
    """
    return BencodeValues(self._decoded_values)

  def IsEmpty(self):
    """Determines if the bencode file has no values (is empty).

    Returns:
      bool: True if the bencode file is empty, False otherwise.
    """
    return not self._decoded_values

  def Open(self, file_object):
    """Opens a bencode file.

    Args:
      file_object (dfvfs.FileIO): file-like object.

    Raises:
      OSError: if the file-like object cannot be read.
      ValueError: if the file-like object is missing.
    """
    if not file_object:
      raise ValueError('Missing file object.')

    file_object.seek(0, os.SEEK_SET)

    try:
      self._decoded_values = bencode.bread(file_object)
    except bencode.BencodeDecodeError as exception:
      raise OSError(exception)

    self._key_names = set()
    for key in self._decoded_values.keys():
      if isinstance(key, bytes):
        # Work-around for issue in bencode 3.0.1 where keys are bytes.
        key = key.decode('utf-8')

      self._key_names.add(key)


class BencodeParser(interface.FileObjectParser):
  """Parser for bencoded files."""

  NAME = 'bencode'
  DATA_FORMAT = 'Bencoded file'

  # Regex match for a bencode dictionary followed by a field size.
  _BENCODE_RE = re.compile(b'd[0-9]')

  _plugin_classes = {}

  def ParseFileObject(self, parser_mediator, file_object):
    """Parses a bencoded file-like object.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      file_object (dfvfs.FileIO): a file-like object.

    Raises:
      WrongParser: when the file cannot be parsed.
    """
    header_data = file_object.read(2)
    if not self._BENCODE_RE.match(header_data):
      raise errors.WrongParser('Not a valid Bencoded file.')

    display_name = parser_mediator.GetDisplayName()
    bencode_file = BencodeFile()

    try:
      bencode_file.Open(file_object)
    except OSError as exception:
      raise errors.WrongParser(
          f'[{self.NAME:s}] unable to parse file: {display_name:s} '
          f'with error: {exception!s}')

    if bencode_file.IsEmpty():
      parser_mediator.ProduceExtractionWarning('missing decoded Bencode values')
      return

    try:
      for plugin_name, plugin in self._plugins_per_name.items():
        if parser_mediator.abort:
          break

        profiling_name = '/'.join([self.NAME, plugin.NAME])

        parser_mediator.SampleFormatCheckStartTiming(profiling_name)

        try:
          result = plugin.CheckRequiredKeys(bencode_file)
        finally:
          parser_mediator.SampleFormatCheckStopTiming(profiling_name)

        if not result:
          logger.debug(
              f'Skipped parsing file: {display_name:s} with plugin: '
              f'{plugin_name:s}')
          continue

        logger.debug(
            f'Parsing file: {display_name:s} with plugin: {plugin_name:s}')

        parser_mediator.SampleStartTiming(profiling_name)

        try:
          plugin.UpdateChainAndProcess(
              parser_mediator, bencode_file=bencode_file)

        except Exception as exception:  # pylint: disable=broad-except
          parser_mediator.ProduceExtractionWarning((
              f'plugin: {plugin_name:s} unable to parse Bencode file with '
              f'error: {exception!s}'))

        finally:
          parser_mediator.SampleStopTiming(profiling_name)

    finally:
      bencode_file.Close()


manager.ParsersManager.RegisterParser(BencodeParser)



"""Plist parser plugin for Apple iOS WiFi Known Networks plist files.

The plist contains information about WiFi networks the device has connected to.
"""

from plaso.containers import events
from plaso.parsers import plist
from plaso.parsers.plist_plugins import interface


class IOSWiFiKnownNetworksEventData(events.EventData):
  """Apple iOS WiFi Known Networks event data.

  Attributes:
    added_time (dfdatetime.DateTimeValues): date the network was added.
    bssid (str): BSSID of the WiFi network.
    channel (int): Channel used by the WiFi network.
    last_associated_time (dfdatetime.DateTimeValues): date the network was last
        associated.
    ssid (str): SSID of the WiFi network.
  """

  DATA_TYPE = 'ios:wifi:known_networks:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(
        data_type=self.DATA_TYPE)
    self.added_time = None
    self.bssid = None
    self.channel = None
    self.last_associated_time = None
    self.ssid = None


class IOSWiFiKnownNetworksPlistPlugin(interface.PlistPlugin):
  """Plist parser plugin for Apple iOS WiFi Known Networks plist files."""

  NAME = 'ios_wifi_known_networks'
  DATA_FORMAT = 'Apple iOS WiFi Known Networks plist file'

  PLIST_PATH_FILTERS = frozenset([
      interface.PlistPathFilter('com.apple.wifi.known-networks.plist')])

  PLIST_KEYS = frozenset([])

  def _ParsePlist(
      self, parser_mediator, match=None, top_level=None, **unused_kwargs):
    """Extract WiFi known network entries.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      match (Optional[dict[str: object]]): keys extracted from PLIST_KEYS.
      top_level (Optional[dict[str: object]]): entire plist file.
    """
    for network_values in top_level.values():
      event_data = IOSWiFiKnownNetworksEventData()
      event_data.added_time = self._GetDateTimeValueFromPlistKey(
          network_values, 'AddedAt')
      # TODO: add support for JoinedByUserAt
      # TODO: add support for JoinedBySystemAt
      # TODO: add support for UpdatedAt
      event_data.ssid = network_values.get('SSID').decode('utf8')

      for bssid_data in network_values.get('BSSList', []):
        event_data.bssid = bssid_data.get('BSSID')
        event_data.channel = bssid_data.get('Channel')
        event_data.last_associated_time = self._GetDateTimeValueFromPlistKey(
            bssid_data, 'LastAssociatedAt')

        parser_mediator.ProduceEventData(event_data)

      # TODO: add support for __OSSpecific__ knownBSSUpdatedDate,
      # prevJoined and WiFiNetworkPasswordModificationDate


plist.PlistParser.RegisterPlugin(IOSWiFiKnownNetworksPlistPlugin)



"""Text parser plugin for Google Drive Sync log files."""

import pyparsing

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.lib import errors
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class GoogleDriveSyncLogEventData(events.EventData):
  """Google Drive Sync log event data.

  Attributes:
    added_time (dfdatetime.DateTimeValues): date and time the log entry
        was added.
    level (str): logging level of event such as "DEBUG", "WARN", "INFO" and
        "ERROR".
    message (str): log message.
    process_identifier (int): process identifier of process which logged event.
    source_code (str): filename:line_number of source file which logged event.
    thread (str): colon-separated thread identifier in the form "ID:name"
        which logged event.
  """

  DATA_TYPE = 'google_drive_sync_log:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.added_time = None
    self.level = None
    self.message = None
    self.process_identifier = None
    self.source_code = None
    self.thread = None


class GoogleDriveSyncLogTextPlugin(interface.TextPluginWithLineContinuation):
  """Text parser plugin for Google Drive Sync log files."""

  NAME = 'gdrive_synclog'
  DATA_FORMAT = 'Google Drive Sync log file'

  ENCODING = 'utf-8'

  _INTEGER = pyparsing.Word(pyparsing.nums).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  # Date and time values are formatted as:
  # 2018-03-01 12:48:14,224 -0800
  _DATE_TIME = pyparsing.Regex(
      r'(?P<date_time>[0-9]{4}-[0-9]{2}-[0-9]{2} '
      r'[0-9]{2}:[0-9]{2}:[0-9]{2}[,.][0-9]{3} [+-][0-9]{4}) ')

  _PROCESS_IDENTIFIER = (
      pyparsing.Suppress('pid=') +
      _INTEGER.set_results_name('process_identifier'))

  _THREAD = pyparsing.Combine(
      pyparsing.Word(pyparsing.nums) + pyparsing.Literal(':') +
      pyparsing.Word(pyparsing.printables))

  _END_OF_LINE = pyparsing.Suppress(pyparsing.LineEnd())

  _LOG_LINE_START = (
      _DATE_TIME +
      pyparsing.Word(pyparsing.alphas).set_results_name('level') +
      _PROCESS_IDENTIFIER +
      # TODO: consider stripping thread identifier/cleaning up thread name?
      _THREAD.set_results_name('thread') +
      pyparsing.Word(pyparsing.printables).set_results_name('source_code'))

  _LOG_LINE = (
      _LOG_LINE_START + pyparsing.restOfLine().set_results_name('body') +
      _END_OF_LINE)

  _LINE_STRUCTURES = [('log_line', _LOG_LINE)]

  # Using a regular expression here is faster on non-match than the log line
  # grammar.
  VERIFICATION_GRAMMAR = pyparsing.Regex(
      r'(?P<date_time>[0-9]{4}-[0-9]{2}-[0-9]{2} '
      r'[0-9]{2}:[0-9]{2}:[0-9]{2}[,.][0-9]{3} [+-][0-9]{4}) '
      r'[A-Z]+ pid=[0-9]+ [0-9]+:\S+[ ]+\S+:[0-9]+ .*\n')

  VERIFICATION_LITERALS = [' ERROR ', ' FATAL ', ' INFO ', ' pid=', ' WARNING ']

  def __init__(self):
    """Initializes a text parser plugin."""
    super().__init__()
    self._body_lines = None
    self._event_data = None

  def _ParseFinalize(self, parser_mediator):
    """Finalizes parsing.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
    """
    if self._event_data:
      self._event_data.message = ' '.join(self._body_lines)
      self._body_lines = None

      parser_mediator.ProduceEventData(self._event_data)
      self._event_data = None

  def _ParseLogline(self, structure):
    """Parses a log line.

    Args:
      structure (pyparsing.ParseResults): structure of tokens derived from
          a line of a text file.
    """
    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    body = self._GetValueFromStructure(structure, 'body', default_value='')
    body = body.strip()

    event_data = GoogleDriveSyncLogEventData()
    event_data.added_time = self._ParseTimeElements(time_elements_structure)
    event_data.level = self._GetValueFromStructure(structure, 'level')
    event_data.process_identifier = self._GetValueFromStructure(
        structure, 'process_identifier')
    event_data.thread = self._GetValueFromStructure(structure, 'thread')
    event_data.source_code = self._GetValueFromStructure(
        structure, 'source_code')

    self._event_data = event_data
    self._body_lines = [body]

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    if key == '_line_continuation':
      body = structure.replace('\n', ' ').strip()
      self._body_lines.append(body)

    else:
      if self._event_data:
        self._event_data.message = ' '.join(self._body_lines)

        parser_mediator.ProduceEventData(self._event_data)

      self._ParseLogline(structure)

  def _ParseTimeElements(self, time_elements_structure):
    """Parses date and time elements of a log line.

    Args:
      time_elements_structure (pyparsing.ParseResults): date and time elements
          of a log line.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived from
          the time elements.
    """
    try:
      date_time_string = time_elements_structure[:23].replace(',', '.')

      date_time = dfdatetime_time_elements.TimeElementsInMilliseconds()
      date_time.CopyFromDateTimeString(date_time_string)

      time_zone_offset = int(time_elements_structure[-4:-2], 10) * 60
      time_zone_offset += int(time_elements_structure[-2:], 10)
      if time_elements_structure[-5] == '-':
        time_zone_offset *= -1

      date_time.time_zone_offset = time_zone_offset
      date_time.is_local_time = False

      return date_time

    except (IndexError, TypeError, ValueError) as exception:
      raise errors.ParseError(
          f'Unable to parse time elements with error: {exception!s}')

  def _ResetState(self):
    """Resets stored values."""
    self._body_lines = None
    self._event_data = None

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the parser.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    try:
      structure = self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    try:
      self._ParseTimeElements(time_elements_structure)
    except errors.ParseError:
      return False

    self._ResetState()

    return True


text_parser.TextLogParser.RegisterPlugin(GoogleDriveSyncLogTextPlugin)



"""Plist parser plugin for MacOS launchd plist files."""

from plaso.containers import events
from plaso.parsers import plist
from plaso.parsers.plist_plugins import interface


class MacOSLaunchdEventData(events.EventData):
  """MacOS launchd event data.

  Attributes:
    name (str): name.
    group_name (str): name of the group.
    program (str): program and arguments.
    user_name (str): name of the user.
  """

  DATA_TYPE = 'macos:launchd:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.name = None
    self.group_name = None
    self.program = None
    self.user_name = None


class MacOSLaunchdPlistPlugin(interface.PlistPlugin):
  """Plist parser plugin for MacOS launchd plist files.

  Further details about fields within the key:
    Label:
      the required key for uniquely identifying the launchd service.
    Program:
      absolute path to the executable. required in the absence of the
      ProgramArguments key.
    ProgramArguments:
      command-line flags for the executable. required in the absence of the
      Program key.
    UserName:
      the job run as the specified user.
    GroupName:
      the job run as the specified group.
  """

  NAME = 'launchd_plist'
  DATA_FORMAT = 'Launchd plist file'

  # The PLIST_PATH is dynamic, the prefix filename is, by default, named using
  # reverse-domain notation. For example, Chrome is com.google.chrome.plist.
  # /System/Library/LaunchDaemons/*.plist
  # /System/Library/LaunchAgents/*.plist
  # /Library/LaunchDaemons/*.plist
  # /Library/LaunchAgents/*.plist
  # ~/Library/LaunchAgents

  # The Mac OS documentation indicates that Label and # ProgramArguments are
  # required keys, lauchd plists have been observed  that contain Label and
  # Program keys.

  PLIST_KEYS = frozenset(['Label'])

  def CheckRequiredFormat(self, top_level):
    """Check if the plist has the minimal structure required by the plugin.

    Args:
      top_level (dict[str, object]): plist top-level item.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    if not super().CheckRequiredFormat(top_level):
      return False

    return 'Program' in top_level or 'ProgramArguments' in top_level

  # pylint: disable=arguments-differ
  def _ParsePlist(self, parser_mediator, top_level=None, **unused_kwargs):
    """Extracts launchd information from the plist.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
        and other components, such as storage and dfVFS.
      top_level (Optional[dict[str, object]]): plist top-level item.
    """
    program = top_level.get('Program')
    program_arguments = top_level.get('ProgramArguments')
    if not (program or program_arguments):
      return
    if program and program_arguments:
      program = ' '.join([program, ' '.join(program_arguments)])
    elif program_arguments:
      program = ' '.join(program_arguments)

    event_data = MacOSLaunchdEventData()
    event_data.group_name = top_level.get('GroupName')
    event_data.name = top_level.get('Label')
    event_data.program = program
    event_data.user_name = top_level.get('UserName')

    parser_mediator.ProduceEventData(event_data)


plist.PlistParser.RegisterPlugin(MacOSLaunchdPlistPlugin)
